import pandas as pd

from sosia.utils import config, flat_set_from_df
from sosia.cache import cache_connect, insert_temporary_table


cache_file = config.get('Cache', 'File path')


def author_cits_in_cache(df, file=cache_file):
    """Search authors citations in cache.

    Parameters
    ----------
    df : DataFrame
        DataFrame of authors to search in a year.

    file : file (optional, default=cache_file)
        The cache file to connect to.

    Returns
    -------
    incache : DataFrame
        DataFrame of results found in cache.

    tosearch: pd.DataFrame
        List of authors not in cache.
    """
    cols = ["auth_id", "year"]
    insert_temporary_table(df, merge_cols=cols, file=file)
    incache = temporary_merge(df, "author_cits_size", merge_cols=cols,
                              file=file)
    if not incache.empty:
        df = df.set_index(cols)
        incache = incache.set_index(cols)
        tosearch = df[~(df.index.isin(incache.index))]
        incache = incache.reset_index()
        tosearch = tosearch.reset_index()
    else:
        tosearch = df
    return incache, tosearch


def authors_in_cache(df, file=cache_file):
    """Search authors in cache.

    Parameters
    ----------
    df : DataFrame
        DataFrame of authors to search.

    file : file (optional, default=cache_file)
        The cache file to connect to.

    Returns
    -------
    incache : DataFrame
        DataFrame of results found in cache.

    tosearch: list
        List of authors not in cache.
    """
    cols = ["auth_id"]
    insert_temporary_table(df, merge_cols=cols, file=file)
    incache = temporary_merge(df, "authors", merge_cols=cols, file=file)
    tosearch = df.auth_id.tolist()
    if not incache.empty:
        incache_list = incache.auth_id.tolist()
        tosearch = [int(au) for au in tosearch if int(au) not in incache_list]
    return incache, tosearch


def author_year_in_cache(df, file=cache_file):
    """Search authors publication information up to year of event in cache.

    Parameters
    ----------
    df : DataFrame
        DataFrame of authors to search with year of the event as second column.

    file : file (optional, default=cache_file)
        The cache file to connect to.

    Returns
    -------
    incache : DataFrame
        DataFrame of results found in cache.

    tosearch: DataFrame
        DataFrame of authors not in cache with year of the event as second
        column.
    """
    cols = ["auth_id", "year"]
    insert_temporary_table(df, merge_cols=cols, file=file)
    incache = temporary_merge(df, "author_year", merge_cols=cols, file=file)
    if not incache.empty:
        df = df.set_index(cols)
        incache = incache.set_index(cols)
        tosearch = df[~(df.index.isin(incache.index))]
        incache = incache.reset_index()
        tosearch = tosearch.reset_index()
        if tosearch.empty:
            cols = ["auth_id", "year", "n_pubs", "n_coauth", "first_year"]
            tosearch = pd.DataFrame(columns=cols)
    else:
        tosearch = df
    return incache, tosearch


def author_size_in_cache(df, file=cache_file):
    """Search authors publication information up to year of event in cache.

    Parameters
    ----------
    df : DataFrame
        DataFrame of authors to search with year of the event as second column.

    file : file (optional, default=cache_file)
        The cache file to connect to.

    Returns
    -------
    incache : DataFrame
        DataFrame of results found in cache.
    """
    cols = ["auth_id", "year"]
    insert_temporary_table(df, merge_cols=cols, file=file)
    incache = temporary_merge(df, "author_size", merge_cols=cols, file=file)
    if incache.empty:
        incache = pd.DataFrame()
    return incache


def sources_in_cache(tosearch, refresh=False, file=cache_file, afid=False):
    """Search sources by year in cache.

    Parameters
    ----------
    tosearch : DataFrame
        DataFrame of sources and years combinations to search.

    refresh : bool (optional, default=False)
        Whether to refresh cached search files.

    file : file (optional, default=cache_file)
        The cache file to connect to.

    afid : bool (optional, default=False)
        If True, search in sources_afids table.

    Returns
    -------
    incache : DataFrame
        DataFrame of results found in cache.

    tosearch: DataFrame
        DataFrame of sources and years combinations not in cache.
    """
    cols = ["source_id", "year"]
    insert_temporary_table(tosearch, merge_cols=cols, file=file)
    c, conn = cache_connect(file=file)
    table = "sources"
    select = "a.source_id, a.year, b.auids"
    if afid:
        table = "sources_afids"
        select = "a.source_id, a.year, b.afid, b.auids"
    q = """SELECT {} FROM temp AS a
        INNER JOIN {} AS b ON a.source_id=b.source_id
        AND a.year=b.year;""".format(select, table)
    incache = pd.read_sql_query(q, conn)
    if not incache.empty:
        incache["auids"] = incache["auids"].str.split(",")
        tosearch = tosearch.merge(incache, "left", on=cols, indicator=True)
        tosearch = tosearch[tosearch["_merge"] == "left_only"].drop("_merge", axis=1)
        tosearch = tosearch[["source_id", "year"]]
        if refresh:
            auth_incache = pd.DataFrame(flat_set_from_df(incache, "auids"),
                                        columns=["auth_id"], dtype="uint64")
            tables = ("authors", "author_size", "author_cits_size", "author_year")
            for table in tables:
                q = "DELETE FROM {} WHERE auth_id=?".format(table)
                conn.executemany(q, auth_incache.to_records(index=False))
                conn.commit()
            tables = ("sources", "sources_afids")
            for table in tables:
                q = "DELETE FROM {} WHERE source_id=? AND year=?".format(table)
                conn.executemany(q, tosearch.to_records(index=False))
                conn.commit()
            incache = pd.DataFrame(columns=cols)
    if tosearch.empty:
        tosearch = pd.DataFrame(columns=cols)
    return incache, tosearch


def temporary_merge(df, table, merge_cols, file):
    """Perform merge with temp table and `table` and retrieve all columns."""
    c, conn = cache_connect(file=file)
    conditions = " and ".join(["a.{0}=b.{0}".format(c) for c in merge_cols])
    q = "SELECT b.* FROM temp AS a INNER JOIN {} AS b ON {};".format(
        table, conditions)
    return pd.read_sql_query(q, conn)



from collections import Counter
from itertools import product

from pandas import DataFrame

from sosia.cache import cache_insert, author_size_in_cache, sources_in_cache
from sosia.processing.queries import base_query, query_journal, query_year
from sosia.utils import custom_print, flat_set_from_df, margin_range,\
    print_progress


def filter_pub_counts(group, ybefore, yupto, npapers, yfrom=None,
                      verbose=False):
    """Filter authors based on restrictions in the number of
    publications in different periods, searched by query_size.

    Parameters
    ----------
    group : list of str
        Scopus IDs of authors to be filtered.

    ybefore : int
        Year to be used as first year. Publications on this year and before
        need to be 0.

    yupto : int
        Year up to which to count publications.

    npapers : list
        List of count of publications, minimum and maximum.

    yfrom : int (optional, default=None)
        If provided, publications are counted only after this year.
        Publications are still set to 0 before ybefore.

    Returns
    -------
    group : list of str
        Scopus IDs filtered.

    pubs_counts : list of int
        List of count of publications within the period provided for authors
        in group.

    older_authors : list of str
        Scopus IDs filtered out because have publications before ybefore.

    Notes
    -----
    It uses cached values first, and searches for more data if needed.
    """
    group = [int(x) for x in group]
    years_check = [ybefore, yupto]
    if yfrom:
        years_check.extend([yfrom - 1])
    authors = DataFrame(list(product(group, years_check)),
                        columns=["auth_id", "year"], dtype="int64")
    authors_size = author_size_in_cache(authors)
    au_skip = []
    group_tocheck = [x for x in group]
    older_authors = []
    pubs_counts = []
    # use information in cache
    if not authors_size.empty:
        # authors that can be already removed because older
        mask = ((authors_size.year <= ybefore) & (authors_size.n_pubs > 0))
        remove = (authors_size[mask]["auth_id"].drop_duplicates().tolist())
        older_authors.extend(remove)
        au_remove = [x for x in remove]
        # remove if number of pubs in year is in any case too small
        mask = ((authors_size.year >= yupto) &
                (authors_size.n_pubs < min(npapers)))
        remove = (authors_size[mask]["auth_id"].drop_duplicates().tolist())
        au_remove.extend(remove)
        # authors with no pubs before min year
        mask = (((authors_size.year == ybefore) & (authors_size.n_pubs == 0)))
        au_ok_miny = (authors_size[mask]["auth_id"].drop_duplicates().tolist())
        # check publications in range
        if yfrom:
            # adjust count by substracting the count before period; keep
            # only authors for which it is possible
            mask = (authors_size.year == yfrom-1)
            authors_size_bef = authors_size[mask]
            authors_size_bef["year"] = yupto
            authors_size_bef.columns = ["auth_id", "year", "n_pubs_bef"]
            bef_auth = set(authors_size_bef["auth_id"])
            mask = ((authors_size["auth_id"].isin(bef_auth)) &
                    (authors_size["year"] == yupto))
            authors_size = authors_size[mask]
            authors_size = authors_size.merge(authors_size_bef, "left",
                                              on=["auth_id", "year"])
            authors_size = authors_size.fillna(0)
            authors_size["n_pubs"] -= authors_size["n_pubs_bef"]
        # authors that can be already removed because of pubs count
        mask = (((authors_size.year >= yupto) &
                 (authors_size.n_pubs < min(npapers))) |
                ((authors_size.year <= yupto) &
                 (authors_size.n_pubs > max(npapers))))
        remove = (authors_size[mask]["auth_id"].drop_duplicates().tolist())
        au_remove.extend(remove)
        # authors with pubs count within the range before the given year
        mask = (((authors_size.year == yupto) &
                 (authors_size.n_pubs >= min(npapers))) &
                (authors_size.n_pubs <= max(npapers)))
        au_ok_year = authors_size[mask][["auth_id", "n_pubs"]].drop_duplicates()
        # authors ok (match both conditions)
        au_ok = list(set(au_ok_miny).intersection(set(au_ok_year["auth_id"])))
        mask = au_ok_year["auth_id"].isin(au_ok)
        pubs_counts = au_ok_year[mask]["n_pubs"].tolist()
        # authors that match only the first condition, but the second is
        # not known, can skip the first cindition check.
        au_skip = [x for x in au_ok_miny if x not in au_remove + au_ok]
        group = [x for x in group if x not in au_remove]
        group_tocheck = [x for x in group if x not in au_skip + au_ok]
    text = "Left with {} authors based on size information already in "\
           "cache.\n{} to check\n".format(len(group), len(group_tocheck))
    custom_print(text, verbose)
    # Verify that publications before minimum year are 0
    if group_tocheck:
        text = "Searching through characteristics of {:,} authors...".format(
            len(group_tocheck))
        custom_print(text, verbose)
        print_progress(0, len(group_tocheck), verbose)
        to_loop = [x for x in group_tocheck]  # Temporary copy
        for i, au in enumerate(to_loop):
            q = "AU-ID({}) AND PUBYEAR BEF {}".format(au, ybefore + 1)
            size = base_query("docs", q, size_only=True)
            tp = (au, ybefore, size)
            cache_insert(tp, table="author_size")
            print_progress(i+1, len(to_loop), verbose)
            if not size == 0:
                group.remove(au)
                group_tocheck.remove(au)
                older_authors.append(au)
        text = "Left with {} authors based on size information before "\
               "minium year\n Filtering based on size query before "\
               "provided year\n".format(len(group))
        custom_print(text, verbose)
    # Verify that publications before the given year falle in range
    group_tocheck.extend(au_skip)
    n = len(group_tocheck)
    if group_tocheck:
        text = "Searching through characteristics of {:,} authors".format(n)
        custom_print(text, verbose)
        print_progress(0, n, verbose)
        for i, au in enumerate(group_tocheck):
            q = "AU-ID({}) AND PUBYEAR BEF {}".format(au, yupto+1)
            n_pubs_yupto = base_query("docs", q, size_only=True)
            tp = (au, yupto, n_pubs_yupto)
            cache_insert(tp, table="author_size")
            # Eventually decrease publication count
            if yfrom and n_pubs_yupto >= min(npapers):
                q = "AU-ID({}) AND PUBYEAR BEF {}".format(au, yfrom)
                n_pubs_yfrom = base_query("docs", q, size_only=True)
                tp = (au, yfrom-1, n_pubs_yfrom)
                cache_insert(tp, table="author_size")
                n_pubs_yupto -= n_pubs_yfrom
            if n_pubs_yupto < min(npapers) or n_pubs_yupto > max(npapers):
                group.remove(au)
            else:
                pubs_counts.append(n_pubs_yupto)
            print_progress(i+1, n, verbose)
    return group, pubs_counts, older_authors


def search_group_from_sources(self, stacked, verbose, refresh=False):
    """Define groups of authors based on publications from a set of sources.

    Parameters
    ----------
    self : sosia.Original
        The object of the Scientist to search information for.

    verbose : bool (optional, default=False)
        Whether to report on the progress of the process.

    refresh : bool (optional, default=False)
        Whether to refresh cached search files.

    Returns
    -------
    today, then, negative : set
        Set of authors publishing in three periods: During the year of
        treatment, during years to match on, and during years before the
        first publication.
    """
    # Filtering variables
    min_year = self.first_year - self.year_margin
    max_year = self.first_year + self.year_margin
    if self.period:
        _margin_setter = self.publications_period
    else:
        _margin_setter = self.publications
    max_pubs = max(margin_range(len(_margin_setter), self.pub_margin))
    years = list(range(min_year, max_year+1))
    search_years = [min_year-1]
    if not self._ignore_first_id:
        search_years.extend(range(min_year, max_year+1))
    search_sources, _ = zip(*self.search_sources)

    # Verbose variables
    n = len(search_sources)
    text = "Searching authors for search_group in {} sources...".format(n)
    custom_print(text, verbose)
    today = set()
    then = set()
    negative = set()

    if stacked:  # Make use of SQL cache
        # Year provided (select also based on location)
        # Get already cached sources from cache
        sources_ay = DataFrame(list(product(search_sources, [self.active_year])),
                               columns=["source_id", "year"])
        _, _search = sources_in_cache(sources_ay, refresh=refresh, afid=True)
        res = query_year(self.active_year, _search.source_id.tolist(), refresh,
                         verbose, afid=True)
        cache_insert(res, table="sources_afids")
        sources_ay, _ = sources_in_cache(sources_ay, refresh=refresh, afid=True)
        # Authors publishing in provided year and locations
        mask = None
        if self.search_affiliations:
            mask = sources_ay.afid.isin(self.search_affiliations)
        today = flat_set_from_df(sources_ay, "auids", mask)
        # Years before active year
        # Get already cached sources from cache
        sources_ys = DataFrame(list(product(search_sources, search_years)),
                               columns=["source_id", "year"])
        _, sources_ys_search = sources_in_cache(sources_ys, refresh=refresh)
        missing_years = set(sources_ys_search.year.tolist())
        # Eventually add information for missing years to cache
        for y in missing_years:
            mask = sources_ys_search.year == y
            _sources_search = sources_ys_search[mask].source_id.tolist()
            res = query_year(y, _sources_search, refresh, verbose)
            cache_insert(res, table="sources")
        # Get full cache
        sources_ys, _ = sources_in_cache(sources_ys, refresh=False)
        # Authors publishing in year(s) of first publication
        if not self._ignore_first_id:
            mask = sources_ys.year.between(min_year, max_year, inclusive=True)
            then = flat_set_from_df(sources_ys, "auids", mask)
        # Authors with publications before
        mask = sources_ys.year < min_year
        negative = flat_set_from_df(sources_ys, "auids", mask)
    else:
        auth_count = []
        print_progress(0, n, verbose)
        for i, source_id in enumerate(search_sources):
            info = query_journal(source_id, [self.active_year] + years, refresh)
            today.update(info[str(self.active_year)])
            if not self._ignore_first_id:
                for y in years:
                    then.update(info[str(y)])
            for y in range(int(min(info.keys())), min_year):
                negative.update(info[str(y)])
            for y in info:
                if int(y) <= self.active_year:
                    auth_count.extend(info[str(y)])
            print_progress(i+1, n, verbose)
        c = Counter(auth_count)
        negative.update({a for a, npub in c.items() if npub > max_pubs})

    return today, then, negative



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Authors:   Michael E. Rose <michael.ernst.rose@gmail.com>
#            Stefano H. Baruffaldi <ste.baruffaldi@gmail.com>
"""Main class for sosia."""

from string import digits, punctuation, Template
from warnings import warn
import pandas as pd

from sklearn.feature_extraction.stop_words import ENGLISH_STOP_WORDS

from sosia.cache import author_cits_in_cache, author_year_in_cache, cache_insert
from sosia.classes import Scientist
from sosia.filtering import search_group_from_sources, filter_pub_counts
from sosia.processing import count_citations, get_authors, base_query,\
    inform_matches, query_author_data, stacked_query
from sosia.utils import accepts, build_dict, custom_print, margin_range,\
    maybe_add_source_names, print_progress

STOPWORDS = list(ENGLISH_STOP_WORDS)
STOPWORDS.extend(punctuation + digits)


class Original(Scientist):
    @property
    def search_group(self):
        """The set of authors that might be matches to the scientist.  The
        set contains the intersection of all authors publishing in the given
        year as well as authors publishing around the year of first
        publication.  Some authors with too many publications in the given
        year and authors having published too early are removed.

        Notes
        -----
        Property is initiated via .define_search_group().
        """
        try:
            return self._search_group
        except AttributeError:
            return None

    @property
    def search_sources(self):
        """The set of sources (journals, books) comparable to the sources
        the scientist published in until the given year.
        A sources is comparable if is belongs to the scientist's main field
        but not to fields alien to the scientist, and if the types of the
        sources are the same as the types of the sources in the scientist's
        main field where she published in.

        Notes
        -----
        Property is initiated via .define_search_sources().
        """
        try:
            return self._search_sources
        except AttributeError:
            return None

    @search_sources.setter
    @accepts((set, list, tuple))
    def search_sources(self, val):
        self._search_sources = maybe_add_source_names(val, self.source_names)

    def __init__(self, scientist, year, year_margin=1, pub_margin=0.1,
                 cits_margin=0.1, coauth_margin=0.1, period=None, refresh=False,
                 eids=None, search_affiliations=None):
        """Class to represent a scientist for which we want to find a control
        group.

        Parameters
        ----------
        scientist : str, int or list or str or int
            Scopus Author ID, or list of Scopus Author IDs, of the scientist
            you want to find control groups for.

        year : str or numeric
            Year of the event.  Control groups will be matched on trends and
            characteristics of the scientist up to this year.

        year_margin : numeric (optional, default=1)
            Number of years by which the search for authors publishing around
            the year of the focal scientist's year of first publication should
            be extend in both directions.

        pub_margin : numeric (optional, default=0.1)
            The left and right margin for the number of publications to match
            possible matches and the scientist on.  If the value is a float,
            it is interpreted as percentage of the scientists number of
            publications and the resulting value is rounded up.  If the value
            is an integer it is interpreted as fixed number of publications.

        cits_margin : numeric (optional, default=0.1)
            The left and right margin for the number of citations to match
            possible matches and the scientist on.  If the value is a float,
            it is interpreted as percentage of the scientists number of
            publications and the resulting value is rounded up.  If the value
            is an integer it is interpreted as fixed number of citations.

        coauth_margin : numeric (optional, default=0.1)
            The left and right margin for the number of coauthors to match
            possible matches and the scientist on.  If the value is a float,
            it is interpreted as percentage of the scientists number of
            coauthors and the resulting value is rounded up.  If the value
            is an integer it is interpreted as fixed number of coauthors.

        period: int (optional, default=None)
            The period in which to consider publications. If not provided,
            all publications are considered.

        refresh : boolean (optional, default=False)
            Whether to refresh all cached files or not.

        eids : list (optional, default=None)
            A list of scopus EIDs of the publications of the scientist you
            want to find a control for.  If it is provided, the scientist
            properties and the control group are set based on this list of
            publications, instead of the list of publications obtained from
            the Scopus Author ID.

        affiliations : list (optional, default=None)
            A list of scopus affiliation IDs. If provided, sosia searches
            for matches within this affiliation in the year provided.
        """
        # Internal checks
        if not isinstance(year_margin, (int, float)):
            raise Exception("Argument year_margin must be float or integer.")
        if not isinstance(pub_margin, (int, float)):
            raise Exception("Argument pub_margin must be float or integer.")
        if not isinstance(coauth_margin, (int, float)):
            raise Exception("Argument coauth_margin must be float or integer.")

        # Variables
        if not isinstance(scientist, list):
            scientist = [scientist]
        self.identifier = [str(auth_id) for auth_id in scientist]
        self.year = int(year)
        self.year_margin = year_margin
        self.pub_margin = pub_margin
        self.cits_margin = cits_margin
        self.coauth_margin = coauth_margin
        self.period = period
        self.eids = eids
        if isinstance(search_affiliations, (int, str)):
            search_affiliations = [search_affiliations]
        if search_affiliations:
            search_affiliations = [int(a) for a in search_affiliations]
        self.search_affiliations = search_affiliations
        self.refresh = refresh

        # Instantiate superclass to load private variables
        Scientist.__init__(self, self.identifier, year, refresh, period)

    def define_search_group(self, stacked=False, verbose=False, refresh=False,
                            ignore_first_id=False):
        """Define search_group.

        Parameters
        ----------
        stacked : bool (optional, default=False)
            Whether to combine searches in few queries or not.  Cached
            files with most likely not be resuable.  Set to True if you
            query in distinct fields or you want to minimize API key usage.

        verbose : bool (optional, default=False)
            Whether to report on the progress of the process.

        refresh : bool (optional, default=False)
            Whether to refresh cached search files.

        ignore_first_id: boolean (optional, default=False)
            If True, the authors in the first year of publication of the
            scientist are not selected based on their Author ID but based on
            their surname and first name.
        """
        # Checks
        if not self.search_sources:
            text = "No search sources defined.  Please run "\
                   ".define_search_sources() first."
            raise Exception(text)
        self._ignore_first_id = ignore_first_id
        if ignore_first_id and not self.period:
            self._ignore_first_id = False
            warn("ignore_first_id set back to False: period is None or "
                 "the first year of the period is before the first year "
                 "of publication of the scientist.")

        # Query journals
        params = {"self": self, "stacked": stacked,
                  "refresh": refresh, "verbose": verbose}
        today, then, negative = search_group_from_sources(**params)

        # Finalize and select
        group = today
        if not self._ignore_first_id:
            group = today.intersection(then)
        negative.update({str(i) for i in self.identifier})
        negative.update({str(i) for i in self.coauthors})
        self._search_group = sorted(list(group - negative))
        text = "Found {:,} authors for search_group".format(len(self._search_group))
        custom_print(text, verbose)
        return self

    def define_search_sources(self, verbose=False):
        """Define .search_sources.

        Parameters
        ----------
        verbose : bool (optional, default=False)
            Whether to report on the progress of the process.
        """
        # Get list of source IDs of scientist's own sources
        own_source_ids, _ = zip(*self.sources)
        # Select sources in scientist's main field
        df = self.field_source
        same_field = df["asjc"] == self.main_field[0]
        # Select sources of same type as those in scientist's main field
        same_sources = same_field & df["source_id"].isin(own_source_ids)
        main_types = set(df[same_sources]["type"])
        same_type = same_field & df["type"].isin(main_types)
        source_ids = df[same_type]["source_id"].unique()
        selected = df[df["source_id"].isin(source_ids)].copy()
        selected["asjc"] = selected["asjc"].astype(str) + " "
        grouped = selected.groupby("source_id").sum()["asjc"].to_frame()
        # Deselect sources with alien fields
        mask = grouped["asjc"].apply(
            lambda s: any(x for x in s.split() if int(x) not in self.fields))
        grouped = grouped[~mask]
        sources = set((s, self.source_names.get(s)) for s in grouped.index)
        # Add own sources
        sources.update(self.sources)
        # Finalize
        self._search_sources = sorted(list(sources))
        types = "; ".join(list(main_types))
        text = "Found {} sources matching main field {} and type(s) {}".format(
            len(self._search_sources), self.main_field[0], types)
        custom_print(text, verbose)
        return self

    def find_matches(self, stacked=False, verbose=False, stop_words=STOPWORDS,
                     information=True, refresh=False, **tfidf_kwds):
        """Find matches within search_group based on four criteria:
        1. Started publishing in about the same year
        2. Has about the same number of publications in the year of treatment
        3. Has about the same number of coauthors in the year of treatment
        4. Has about the same number of citations in the year of treatment
        5. Works in the same field as the scientist's main field

        Parameters
        ----------
        stacked : bool (optional, default=False)
            Whether to combine searches in few queries or not.  Cached
            files will most likely not be resuable.  Set to True if you
            query in distinct fields or you want to minimize API key usage.

        verbose : bool (optional, default=False)
            Whether to report on the progress of the process.

        stop_words : list (optional, default=STOPWORDS)
            A list of words that should be filtered in the analysis of
            abstracts.  Default list is the list of english stopwords
            by nltk, augmented with numbers and interpunctuation.

        information : bool or iterable (optional, default=True)
            Whether to return additional information on the matches that may
            help in the selection process.  If an iterable of keywords is
            provied, only return information for these keywords.  Allowed
            values are "first_year", "num_coauthors", "num_publications",
            "num_citations", "country", "language",
            "reference_sim", "abstract_sim".

        refresh : bool (optional, default=False)
            Whether to refresh cached search files.

        tfidf_kwds : keywords
            Parameters to pass to TfidfVectorizer from the sklearn package
            for abstract vectorization.  Not used when `information=False` or
            or when "abstract_sim" is not in `information`.  See
            https://scikit-learn.org/stable/modules/generated/sklearn.feature_extraction.text.TfidfVectorizer.html
            for possible values.

        Returns
        -------
        matches : list
            A list of Scopus IDs of scientists matching all the criteria (if
            information is False) or a list of namedtuples with the Scopus ID
            and additional information (if information is True).

        Raises
        ------
        ValueError
            If information is not bool and contains invalid keywords.
        """
        # Checks
        info_keys = ["first_name", "surname", "first_year", "num_coauthors",
                     "num_publications", "num_citations", "num_coauthors_period",
                     "num_publications_period", "num_citations_period",
                     "subjects", "country", "affiliation_id", "affiliation",
                     "language", "reference_sim", "abstract_sim"]
        if isinstance(information, bool):
            if information:
                keywords = info_keys
            elif self.search_affiliations:
                information = True
                keywords = ["affiliation_id"]
            else:
                keywords = None
        else:
            keywords = information
            invalid = [x for x in keywords if x not in info_keys]
            if invalid:
                text = ("Parameter information contains invalid keywords: ",
                        ", ".join(invalid))
                raise ValueError(text)
            if self.search_affiliations and "affiliation_id" not in keywords:
                keywords.append("affiliation_id")
        # Variables
        _years = range(self.first_year-self.year_margin,
                       self.first_year+self.year_margin+1)
        if self.period:
            _npapers = margin_range(len(self.publications_period), self.pub_margin)
            _ncits = margin_range(self.citations_period, self.cits_margin)
            _ncoauth = margin_range(len(self.coauthors_period), self.coauth_margin)
            _npapers_full = margin_range(len(self.publications), self.pub_margin)
            _ncits_full = margin_range(self.citations, self.cits_margin)
            _ncoauth_full = margin_range(len(self.coauthors), self.coauth_margin)
        else:
            _npapers = margin_range(len(self.publications), self.pub_margin)
            _ncits = margin_range(self.citations, self.cits_margin)
            _ncoauth = margin_range(len(self.coauthors), self.coauth_margin)
        n = len(self.search_group)
        text = "Searching through characteristics of {:,} authors".format(n)
        custom_print(text, verbose)

        # First round of filtering: minimum publications and main field
        # create df of authors
        authors = query_author_data(self.search_group, verbose=verbose)
        same_field = (authors.areas.str.startswith(self.main_field[1]))
        enough_pubs = (authors.documents.astype(int) >= int(min(_npapers)))
        group = authors[same_field & enough_pubs]["auth_id"].tolist()
        group.sort()
        n = len(group)
        text = "Left with {} authors\nFiltering based on provided "\
               "conditions...".format(n)
        custom_print(text, verbose)

        # Second round of filtering:
        # Check having no publications before minimum year, and if 0, the
        # number of publications in the relevant period.
        params = {"group": group, "ybefore": min(_years)-1,
                  "yupto": self.year, "npapers": _npapers,
                  "yfrom": self.year_period, "verbose": verbose}
        group, _, _ = filter_pub_counts(**params)
        # Also screen out ids with too many publications over the full period
        if self.period:
            params.update({"npapers": [1, max(_npapers_full)], "yfrom": None,
                           "group": group})
            group, _, _ = filter_pub_counts(**params)

        # Third round of filtering: citations (in the FULL period).
        authors = pd.DataFrame({"auth_id": group, "year": self.year})
        _, authors_cits_search = author_cits_in_cache(authors)
        text = "Search and filter based on count of citations\n{} to search "\
               "out of {}\n".format(len(authors_cits_search), len(group))
        custom_print(text, verbose)
        if not authors_cits_search.empty:
            authors_cits_search['n_cits'] = 0
            print_progress(0, len(authors_cits_search), verbose)
            for i, au in authors_cits_search.iterrows():
                q = "REF({}) AND PUBYEAR BEF {} AND NOT AU-ID({})".format(
                    au['auth_id'], self.year + 1, au['auth_id'])
                n = base_query("docs", q, size_only=True)
                authors_cits_search.at[i, 'n_cits'] = n
                print_progress(i + 1, len(authors_cits_search), verbose)
            cache_insert(authors_cits_search, table="author_cits_size")
        auth_cits_incache, _ = author_cits_in_cache(authors[["auth_id", "year"]])
        # keep if citations are in range
        mask = ((auth_cits_incache.n_cits <= max(_ncits)) &
                (auth_cits_incache.n_cits >= min(_ncits)))
        if self.period:
            mask = ((auth_cits_incache.n_cits >= min(_ncits)) &
                    (auth_cits_incache.n_cits <= max(_ncits_full)))
        group = (auth_cits_incache[mask]['auth_id'].tolist())

        # Fourth round of filtering: Download publications, verify coauthors
        # (in the FULL period) and first year.
        n = len(group)
        text = "Left with {} authors\nFiltering based on coauthors "\
               "number...".format(n)
        custom_print(text, verbose)
        authors = pd.DataFrame({"auth_id": group, "year": self.year},
                               dtype="uint64")
        _, author_year_search = author_year_in_cache(authors)
        matches = []
        if stacked:  # Combine searches
            if not author_year_search.empty:
                q = Template("AU-ID($fill) AND PUBYEAR BEF {}".format(
                    self.year + 1))
                auth_year_group = author_year_search.auth_id.tolist()
                params = {"group": auth_year_group, "res": [],
                          "template": q, "refresh": refresh,
                          "joiner": ") OR AU-ID(", "q_type": "docs"}
                if verbose:
                    params.update({"total": len(auth_year_group)})
                res, _ = stacked_query(**params)
                res = build_dict(res, auth_year_group)
                if res:
                    # res can become empty after build_dict if a au_id is old
                    res = pd.DataFrame.from_dict(res, orient="index")
                    res["year"] = self.year
                    res = res[["year", "first_year", "n_pubs", "n_coauth"]]
                    res.index.name = "auth_id"
                    res = res.reset_index()
                    cache_insert(res, table="author_year")
            author_year_cache, _ = author_year_in_cache(authors)
            if self._ignore_first_id:
                # only number of coauthors should be big enough
                enough = (author_year_cache.n_coauth >= min(_ncoauth))
                notoomany = (author_year_cache.n_coauth <= max(_ncoauth_full))
                mask = enough & notoomany
            elif self.period:
                # number of coauthors should be "big enough" and first year in
                # window
                same_start = (author_year_cache.first_year.between(min(_years),
                              max(_years)))
                enough = (author_year_cache.n_coauth >= min(_ncoauth))
                notoomany = (author_year_cache.n_coauth <= max(_ncoauth_full))
                mask = same_start & enough & notoomany
            else:
                # all restrictions apply
                same_start = (author_year_cache.first_year.between(min(_years),
                              max(_years)))
                same_coauths = (author_year_cache.n_coauth.between(min(_ncoauth),
                                max(_ncoauth)))
                mask = same_start & same_coauths
            matches = author_year_cache[mask]["auth_id"].tolist()
        else:  # Query each author individually
            for i, au in enumerate(group):
                print_progress(i + 1, len(group), verbose)
                res = base_query("docs", "AU-ID({})".format(au), refresh=refresh)
                res = [p for p in res if p.coverDate and
                       int(p.coverDate[:4]) <= self.year]
                # Filter
                min_year = int(min([p.coverDate[:4] for p in res]))
                authids = [p.author_ids for p in res if p.author_ids]
                authors = set([a for p in authids for a in p.split(";")])
                n_coauth = len(authors) - 1  # Subtract 1 for focal author
                if self._ignore_first_id and (n_coauth < max(_ncoauth)):
                    # only number of coauthors should be big enough
                    continue
                elif (self.period and ((n_coauth < max(_ncoauth)) or
                      (min_year not in _years))):
                    # number of coauthors should be "big enough" and first year
                    # in window
                    continue
                elif ((len(res) not in _npapers) or (min_year not in _years) or
                        (n_coauth not in _ncoauth)):
                    continue
                matches.append(au)

        if self.period:
            text = "Left with {} authors\nFiltering based on exact period "\
                   "citations and coauthors...".format(len(matches))
            custom_print(text, verbose)
            # Further screen matches based on period cits and coauths
            to_loop = [m for m in matches]  # temporary copy
            for m in to_loop:
                q = "AU-ID({})".format(m)
                res = base_query("docs", "AU-ID({})".format(m), refresh=refresh,
                    fields=["eid", "author_ids", "coverDate"])
                pubs = [p for p in res if int(p.coverDate[:4]) <= self.year and
                        int(p.coverDate[:4]) >= self.year_period]
                coauths = set(get_authors(pubs)) - {str(m)}
                if not (min(_ncoauth) <= len(coauths) <= max(_ncoauth)):
                    matches.remove(m)
                    continue
                eids_period = [p.eid for p in pubs]
                cits = count_citations(search_ids=eids_period,
                        pubyear=self.year+1, exclusion_key="AU-ID",
                        exclusion_ids=[str(m)])
                if not (min(_ncits) <= cits <= max(_ncits)):
                    matches.remove(m)
        text = "Found {:,} author(s) matching all criteria".format(len(matches))
        custom_print(text, verbose)

        # Possibly add information to matches
        if keywords and len(matches) > 0:
            custom_print("Providing additional information...", verbose)
            profiles = [Scientist([str(a)], self.year, period=self.period,
                                  refresh=refresh) for a in matches]
            matches = inform_matches(profiles, self, keywords, stop_words,
                                     verbose, refresh, **tfidf_kwds)
        if self.search_affiliations:
            matches = [m for m in matches if 
                       len(set(m.affiliation_id.replace(" ","").split(";"))
                       .intersection([str(a) for a in
                                      self.search_affiliations]))]
        return matches



import sqlite3
from os import makedirs
from os.path import exists, expanduser
import configparser

import pandas as pd
import requests
from bs4 import BeautifulSoup

from sosia.utils import CACHE_TABLES, CONFIG_FILE, FIELDS_SOURCES_LIST,\
    SOURCES_NAMES_LIST, URL_CONTENT, URL_SOURCES

# Configuration setup
file_exists = exists(CONFIG_FILE)
config = configparser.ConfigParser()
config.optionxform = str
if not file_exists:
    config.add_section('Cache')
    _cache_default = expanduser("~/.sosia/") + "cache_sqlite.sqlite"
    config.set('Cache', 'File path', _cache_default)
    try:
        with open(CONFIG_FILE, 'w') as f:
            config.write(f)
    except FileNotFoundError:  # Fix for sphinx build
        pass
else:
    config.read(CONFIG_FILE)


def create_cache(drop=False, file=None):
    """Create or recreate tables in cache file.

    Parameters
    ----------
    drop : boolean (optional, default=False)
        If True, deletes and recreates all tables in cache (irreversible).

    file : file (optional, default=CACHE_SQLITE)
        The name of the cache file to be used. By default is named
        cache_sqlite.sqlite and located in "~/.sosia/".
    """
    if not file:
        file = config.get("Cache", "File path")
    conn = sqlite3.connect(file)
    c = conn.cursor()
    for table, variables in CACHE_TABLES.items():
        if drop:
            q = "DROP TABLE IF EXISTS {}".format(table)
            c.execute(q)
        columns = ", ".join(" ".join(v) for v in variables["columns"])
        prim_keys = ", ".join(variables["primary"])
        q = "CREATE TABLE IF NOT EXISTS {} ({}, PRIMARY KEY({}))".format(
            table, columns, prim_keys)
        c.execute(q)


def create_fields_sources_list():
    """Download Scopus files with information on covered sources and create
    one list of all sources with ids and one with field information.
    """
    # Set up
    path = expanduser("~/.sosia/")
    if not exists(path):
        makedirs(path)
    rename = {
        "All Science Journal Classification Codes (ASJC)": "asjc",
        "Scopus ASJC Code (Sub-subject Area)": "asjc",
        "ASJC code": "asjc",
        "Source Type": "type",
        "Type": "type",
        "Sourcerecord id": "source_id",
        "Scopus Source ID": "source_id",
        "Scopus SourceID": "source_id",
        "Scopus Source ID": "source_id",
        "Title": "title",
        "Source title": "title",
    }
    keeps = list(set(rename.values()))

    # Get Information from Scopus Sources list
    sources = pd.read_excel(URL_SOURCES, sheet_name=None, header=1)
    _drop_sheets(sources, ["About CiteScore", "ASJC Codes", "Sheet1"])
    out = pd.concat(
        [df.rename(columns=rename)[keeps].dropna() for df in sources.values()]
    )
    out = out.drop_duplicates()

    # Add information from list of external publication titles
    external = pd.read_excel(_get_source_title_url(), sheet_name=None)
    _drop_sheets(external, ["More info Medline", "ASJC classification codes"])

    for df in external.values():
        _update_dict(rename, df.columns, "source title", "title")
        if "Source Type" not in df.columns:
            df["type"] = "conference proceedings"
        subset = df.rename(columns=rename)[keeps].dropna()
        subset["asjc"] = subset["asjc"].astype(str).apply(_clean).str.split()
        subset = (subset.set_index(["source_id", "title", "type"])
                        .asjc.apply(pd.Series)
                        .stack()
                        .rename("asjc")
                        .reset_index()
                        .drop("level_3", axis=1))
        out = pd.concat([out, subset], sort=True)

    # Write list of names
    order = ["source_id", "title"]
    names = out[order].drop_duplicates().sort_values("source_id")
    names.to_csv(SOURCES_NAMES_LIST, index=False)

    # Write list of fields by source
    out["type"] = out["type"].str.lower().str.strip()
    out.drop("title", axis=1).to_csv(FIELDS_SOURCES_LIST, index=False)


def _clean(x):
    """Auxiliary function to clean a string Series."""
    return x.replace(";", " ").replace(",", " ").replace("  ", " ").strip()


def _drop_sheets(sheets, drops):
    """Auxiliary function to drop sheets from an Excel DataFrame."""
    for drop in drops:
        try:
            sheets.pop(drop)
        except KeyError:
            continue


def _get_source_title_url():
    """Extract the link to the most recent Scopus sources list."""
    resp = requests.get(URL_CONTENT)
    soup = BeautifulSoup(resp.text, "lxml")
    try:
        return soup.find("a", {"title": "source list"})["href"]
    except AttributeError:
        raise ValueError("Link to sources list not found.")


def _update_dict(d, lst, key, replacement):
    """Auxiliary function to add keys to a dictionary if a given string is
    included in the key.
    """
    for c in lst:
        if c.lower().startswith(key):
            d[c] = replacement



from pbr.version import VersionInfo

_v = VersionInfo("sosia").semantic_version()
__version__ = _v.release_string()
version_info = _v.version_tuple()

from sosia.sosia import Original
from sosia.utils import *



import pandas as pd
from collections import defaultdict
from string import Template
from time import sleep
from urllib.error import HTTPError

from pybliometrics.scopus import AuthorSearch, ScopusSearch
from pybliometrics.scopus.exception import Scopus400Error, ScopusQueryError,\
    Scopus500Error

from sosia.processing.extraction import get_authors, get_auth_from_df,\
    expand_affiliation
from sosia.utils import print_progress
from sosia.cache import authors_in_cache, cache_insert


def base_query(q_type, query, refresh=False, fields=None, size_only=False):
    """Wrapper function to perform a particular search query.

    Parameters
    ----------
    q_type : str
        Determines the query search that will be used.  Allowed values:
        "author", "docs".

    query : str
        The query string.

    refresh : bool (optional, default=False)
        Whether to refresh cached files if they exist, or not.

    fields : list of field names (optional, default=None)
        Fields in the Scopus query that must always present.  To be passed
         onto pybliometrics.scopus.ScopusSearch.  Will be ignored
         when q_type = "author".

    size_only : bool (optional, default=False)
        Whether to not download results and return the number
        of results instead.

    tsleep: float
        Seconds to wait in case of failure due to errors.

    Returns
    -------
    res : list of namedtuples (if size_only is False) or int
        Documents represented by namedtuples as returned from scopus or
        number of search results.

    Raises
    ------
    ValueError:
        If q_type is none of the allowed values.
    """
    params = {"query": query, "refresh": refresh, "download": not size_only}
    # Download query until server is available
    try:
        if q_type == "author":
            obj = AuthorSearch(**params)
        elif q_type == "docs":
            params["integrity_fields"] = fields
            obj = ScopusSearch(**params)
    except (AttributeError, Scopus500Error, KeyError, HTTPError):
        # exception of all errors here has to be maintained due to the
        # occurrence of not replicable errors (e.g. 'cursor', HTTPError)
        sleep(2.0)
        return base_query(q_type, query, refresh=True, fields=None,
                          size_only=size_only)
    if size_only:
        return obj.get_results_size()
    # Parse results, refresh once if integrity check fails or when server
    # sends bad results (in this case pause querying for a while)
    try:
        if q_type == "author":
            res = obj.authors or []
        elif q_type == "docs":
            res = obj.results or []
    except (AttributeError, Scopus500Error, KeyError, HTTPError):
        # exception of all errors here has to be maintained due to the
        # occurrence of not replicable errors (e.g. 'cursor', HTTPError)
        return base_query(q_type, query, refresh=True, fields=None,
                          size_only=size_only)
    return res


def count_citations(search_ids, pubyear, exclusion_key, exclusion_ids):
    """Auxiliary function to count non-self citations up to a year."""
    q_search_ids = " OR ".join(search_ids)
    joiner = " OR "
    if exclusion_key == "AU-ID":
        joiner = ") AND NOT AU-ID("
    q_exclusion_ids = joiner.join(exclusion_ids)
    s = Template("REF($search_ids) AND PUBYEAR BEF $pubyear AND NOT "
                 "$exclusion_key($exclusion_ids)")
    q = s.substitute(search_ids=q_search_ids, pubyear=pubyear,
                     exclusion_key=exclusion_key, exclusion_ids=q_exclusion_ids)
    # break query if too long
    if len(q) > 3785:
        mid = len(search_ids) // 2
        count1 = count_citations(search_ids[:mid], pubyear, exclusion_key,
                                 exclusion_ids)
        count2 = count_citations(search_ids[mid:], pubyear, exclusion_key,
                                 exclusion_ids)
        return count1 + count2
    return base_query("docs", q, size_only=True)


def query_author_data(authors_list, refresh=False, verbose=False):
    """Wrapper function to search author data for a list of authors, searching
    first in cache and then via stacked search.

    Parameters
    ----------
    authors_list : list
       List of Scopus Author IDs to search.

    refresh : bool (optional, default=False)
        Whether to refresh scopus cached files if they exist, or not.

    verbose : bool (optional)
        Whether to print information on the search progress.

    Returns
    -------
    authors_data : DataFrame
        A dataframe with authors data from AuthorSearch for the list provided.
    """
    authors = pd.DataFrame(authors_list, columns=["auth_id"], dtype="int64")
    # merge existing data in cache and separate missing records
    auth_done, auth_missing = authors_in_cache(authors)
    if auth_missing:
        params = {"group": auth_missing, "res": [],
                  "refresh": refresh, "joiner": ") OR AU-ID(",
                  "q_type": "author", "template": Template("AU-ID($fill)")}
        if verbose:
            print("Pre-filtering...")
            params.update({"total": len(auth_missing)})
        res, _ = stacked_query(**params)
        res = pd.DataFrame(res)
        cache_insert(res, table="authors")
        auth_done, _ = authors_in_cache(authors)
    return auth_done


def query_journal(source_id, years, refresh):
    """Get authors by year for a particular source.

    Parameters
    ----------
    source_id : str or int
        The Scopus ID of the source.

    years : container of int or container of str
        The relevant pulication years to search for.

    refresh : bool (optional)
        Whether to refresh cached files if they exist, or not.

    Returns
    -------
    d : dict
        Dictionary keyed by year listing all authors who published in
        that year.
    """
    try:  # Try complete publication list first
        q = "SOURCE-ID({})".format(source_id)
        if base_query("docs", q, size_only=True) > 5000:
            raise ScopusQueryError()
        res = base_query("docs", q, refresh=refresh)
    except (ScopusQueryError, Scopus500Error):  # Fall back to year-wise queries
        res = []
        for year in years:
            q = Template("SOURCE-ID({}) AND PUBYEAR IS $fill".format(source_id))
            params = {"group": [year], "res": [], "template": q,
                      "joiner": "", "q_type": "docs", "refresh": refresh}
            ext, _ = stacked_query(**params)
            if not valid_results(ext):  # Reload queries with missing years
                params.update({"refresh": True})
                ext, _ = stacked_query(**params)
            res.extend(ext)
    # Sort authors by year
    d = defaultdict(list)
    for pub in res:
        try:
            year = pub.coverDate[:4]
        except TypeError:  # missing year
            continue
        d[year].extend(get_authors([pub]))  # Populate dict
    return d


def query_year(year, source_ids, refresh, verbose, afid=False):
    """Get authors lists for each source in a list and in a particular year.

    Parameters
    ----------
    year : int
        The year of the search.

    source_ids : list
        List of Scopus IDs of sources to search.

    refresh : bool (optional)
        Whether to refresh cached files if they exist, or not.

    verbose : bool (optional)
        Whether to print information on the search progress.

    afid : bool (optional)
        If True, mantains information on the Scopus affiliation ID in res.
    """
    params = {"group": [str(x) for x in sorted(source_ids)],
              "joiner": " OR ", "refresh": refresh, "q_type": "docs"}
    if verbose:
        params.update({"total": len(source_ids)})
        print("Searching authors in {} sources in {}...".format(
            len(source_ids), year))
    q = Template("SOURCE-ID($fill) AND PUBYEAR IS {}".format(year))
    params.update({"template": q, "res": []})
    res, _ = stacked_query(**params)
    res = pd.DataFrame(res)
    if res.empty:
        return res
    res = res.dropna(subset=["author_ids"])
    if res.empty:
        return res
    grouping_cols = ["source_id", "year"]
    if afid:
        res = expand_affiliation(res)
        grouping_cols.append("afid")
    res["year"] = year
    res = (res.astype(str)
              .groupby(grouping_cols)[["author_ids"]]
              .apply(get_auth_from_df)
              .reset_index()
              .rename(columns={0: "auids"}))
    return res

def stacked_query(group, res, template, joiner, q_type, refresh,
                  i=0, total=None):
    """Auxiliary function to recursively perform queries until they work.

    Parameters
    ----------
    group : list of str
        Scopus IDs (of authors or sources) for which the stacked query should
        be conducted.

    res : list
        (Initially empty )Container to which the query results will be
        appended.

    template : Template()
        A string template with one paramter named `fill` which will be used
        as search query.

    joiner : str
        On wich the group elements should be joined to fill the query.

    q_type : str
        Determines the query search that will be used.  Allowed values:
        "author", "docs".

    refresh : bool
        Whether the cached files should be refreshed or not.

    i : int (optional, default=0)
        A count variable to be used for printing the progress bar.

    total : int (optional, default=None)
        The total number of elements in the group.  If provided, a progress
        bar will be printed.

    Returns
    -------
    res : list
        A list of namedtuples representing publications.

    i : int
        A running variable to indicate the progress.

    Notes
    -----
    Results of each successful query are appended to ´res´.
    """
    group = [str(g) for g in group]  # make robust to passing int
    q = template.substitute(fill=joiner.join(group))
    try:
        n = base_query(q_type, q, size_only=True)
        if n > 5000 and len(group) > 1:
            raise ScopusQueryError()
        res.extend(base_query(q_type, q, refresh=refresh))
        verbose = total is not None
        i += len(group)
        print_progress(i, total, verbose)
    except (Scopus400Error, Scopus500Error, ScopusQueryError) as e:
        # Split query group into two equally sized groups
        mid = len(group) // 2
        params = {"group": group[:mid], "res": res, "template": template,
                  "i": i, "joiner": joiner, "q_type": q_type, "total": total,
                  "refresh": refresh}
        res, i = stacked_query(**params)
        params.update({"group": group[mid:], "i": i})
        res, i = stacked_query(**params)
    return res, i


def valid_results(res):
    """Verify that each element ScopusSearch in `res` contains year info."""
    try:
        _ = [p for p in res if p.subtype == "ar" and int(p.coverDate[:4])]
        return True
    except:
        return False



from math import ceil, inf
from collections import Counter, defaultdict

from pandas import read_csv

from sosia.utils.constants import ASJC_2D, FIELDS_SOURCES_LIST,\
    SOURCES_NAMES_LIST
from sosia.utils.startup import create_fields_sources_list


def add_source_names(source_ids, names):
    """Add names of sources to list of source IDs turning the list into a
    list of tuples.
    """
    return sorted([(s_id, names.get(int(s_id))) for s_id in set(source_ids)])


def build_dict(results, chunk):
    """Create dictionary assigning publication information to authors we
    are looking for.
    """
    chunk = [int(au) for au in chunk]
    d = defaultdict(
        lambda: {"first_year": inf, "pubs": set(), "coauth": set(),
                 "n_coauth": inf, "n_pubs": inf})
    for pub in results:
        if not pub.author_ids:
            continue
        authors = set([int(au) for au in pub.author_ids.split(";")])
        for focal in authors.intersection(chunk):
            d[focal]["coauth"].update(authors)
            d[focal]["coauth"].remove(focal)
            d[focal]["pubs"].add(pub.eid)
            d[focal]["n_pubs"] = len(d[focal]["pubs"])
            d[focal]["n_coauth"] = len(d[focal]["coauth"])
            if not pub.coverDate:
                continue
            first_year = min(d[focal]["first_year"], int(pub.coverDate[:4]))
            d[focal]["first_year"] = first_year
    return d


def custom_print(text, verbose):
    """Print text if verbose."""
    if verbose:
        print(text)


def flat_set_from_df(df, col, condition=None):
    """Flatten Series from DataFrame which contains lists and
    return as set, optionally after filtering the DataFrame.
    """
    if condition is not None:
        df = df[condition]
    lists = df[col].tolist()
    return set([item for l in lists for item in l])


def get_main_field(fields):
    """Get main 4-digit ASJC field (code) and main 2-digit ASJC field (name).


    Parameters
    ----------
    fields : iterable of int
        Lists of fields the researcher is active in.

    Returns
    -------
    t : tuple
        A 2-element tuple where the first element is the most common 4-digit
        ASJC field and the second element the short name of the most common
        2-digit ASJC field.

    Note
    ----
    We exclude multidisciplinary and give preference to non-general fields.
    """
    # 4 digit field
    c = Counter(fields)
    try:
        c.pop(1000)  # Exclude Multidisciplinary
    except KeyError:
        pass
    top_fields = [f for f, val in c.items() if val == max(c.values())]
    if len(top_fields) == 1:
        main_4 = top_fields[0]
    else:
        non_general_fields = [f for f in top_fields if f % 1000 != 0]
        if non_general_fields:
            main_4 = non_general_fields[0]
        else:
            main_4 = top_fields[0]
    # 2 digit field
    c = Counter([str(f)[:2] for f in fields])
    try:
        c.pop(10)  # Exclude Multidisciplinary
    except KeyError:
        pass
    main_2 = int(c.most_common(1)[0][0])
    return (main_4, ASJC_2D[main_2])


def margin_range(base, val):
    """Create a range of margins around a base value.

    Parameters
    ----------
    base : int
        The value around which a margin should be created.

    val : int or float
        The margin size.  If float, val will be interpreted as percentage.

    Returns
    -------
    r : range
        A range object representing the margin range.
    """
    if isinstance(val, float):
        margin = ceil(val * base)
        r = range(base - margin, base + margin + 1)
    elif isinstance(val, int):
        r = range(base - val, base + val + 1)
    else:
        raise Exception("Value must be either float or int.")
    return r


def maybe_add_source_names(source_ids, names):
    """Add names of sources to list of source IDs turning the list into a
    list of tuples.
    """
    if isinstance(source_ids[0], tuple):
        return source_ids
    else:
        return add_source_names(source_ids, names)


def print_progress(iteration, total, verbose=True, length=50):
    """Print terminal progress bar."""
    if not verbose:
        return None
    percent = 100 * (iteration / float(total))
    filled_len = int(length * iteration // total)
    bar = "█" * filled_len + "-" * (length - filled_len)
    print("\rProgress: |{}| {:.2f}% Complete".format(bar, percent), end="\r")
    if iteration == total:
        print()


def read_fields_sources_list():
    """Auxiliary function to read FIELDS_SOURCES_LIST and create it before,
    if necessary.
    """
    try:
        sources = read_csv(FIELDS_SOURCES_LIST)
        names = read_csv(SOURCES_NAMES_LIST)
    except FileNotFoundError:
        create_fields_sources_list()
        sources = read_csv(FIELDS_SOURCES_LIST)
        names = read_csv(SOURCES_NAMES_LIST)
    return sources, names


def run(op, *args):
    """Call a function passed by partial()."""
    return op(*args)


def _get_obj_name(obj):
    """Auxiliary function to retrieve the name of an object."""
    name = str(obj).replace("class ", "")
    return name.translate(str.maketrans({c: "" for c in "(<>),'"}))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Author:   Michael E. Rose <michael.ernst.rose@gmail.com>
"""Super class to represent a scientist."""

from pybliometrics.scopus import AbstractRetrieval

from sosia.processing import count_citations, base_query, find_location,\
    get_authors, query_author_data
from sosia.utils import accepts, add_source_names, get_main_field,\
    maybe_add_source_names, read_fields_sources_list


class Scientist(object):
    @property
    def active_year(self):
        """The scientist's year of first publication, as integer."""
        return self._active_year

    @active_year.setter
    @accepts(int)
    def active_year(self, val):
        self._active_year = val

    @property
    def affiliation_id(self):
        """The affiliation ID (as string) of the scientist's most frequent
        affiliation in the most recent year (before the given year) that
        the scientist published.
        """
        return self._affiliation_id

    @affiliation_id.setter
    @accepts(str)
    def affiliation_id(self, val):
        self._affiliation_id = val

    @property
    def citations(self):
        """The citations of the scientist until the given year."""
        return self._citations

    @citations.setter
    @accepts(int)
    def citations(self, val):
        self._citations = val

    @property
    def citations_period(self):
        """The citations of the scientist until during the given period."""
        return self._citations_period

    @citations_period.setter
    @accepts(int)
    def citations_period(self, val):
        self._citations_period = val

    @property
    def city(self):
        """City of the scientist's most frequent affiliation
        in the most recent year (before the given year) that
        the scientist published.
        """
        return self._city

    @city.setter
    @accepts(str)
    def city(self, val):
        self._city = val

    @property
    def country(self):
        """Country of the scientist's most frequent affiliation
        in the most recent year (before the given year) that
        the scientist published.
        """
        return self._country

    @country.setter
    @accepts(str)
    def country(self, val):
        self._country = val

    @property
    def coauthors(self):
        """Set of coauthors of the scientist on all publications until the
        given year.
        """
        return self._coauthors

    @coauthors.setter
    @accepts((set, list, tuple))
    def coauthors(self, val):
        self._coauthors = val

    @property
    def coauthors_period(self):
        """Set of coauthors of the scientist on all publications during the
        given period.
        """
        return self._coauthors_period

    @coauthors_period.setter
    @accepts((set, list, tuple))
    def coauthors_period(self, val):
        self._coauthors_period = val

    @property
    def fields(self):
        """The fields of the scientist until the given year, estimated from
        the sources (journals, books, etc.) she published in.
        """
        return self._fields

    @fields.setter
    @accepts((set, list, tuple))
    def fields(self, val):
        self._fields = val

    @property
    def first_year(self):
        """The scientist's year of first publication, as integer."""
        return self._first_year

    @first_year.setter
    @accepts(int)
    def first_year(self, val):
        self._first_year = val

    @property
    def first_name(self):
        """The scientist's first name."""
        return self._first_name

    @first_name.setter
    @accepts(str)
    def first_name(self, val):
        self._name = val

    @property
    def main_field(self):
        """The scientist's main field of research, as tuple in
        the form (ASJC code, general category).

        The main field is the field with the most publications, provided it
        is not Multidisciplinary (ASJC code 1000).  In case of an equal number
        of publications, preference is given to non-general fields (those
        whose ASJC ends on a digit other than 0).
        """
        return self._main_field

    @main_field.setter
    def main_field(self, val):
        if not isinstance(val, tuple) or len(val) != 2:
            raise Exception("Value must be a two-element tuple.")
        self._main_field = val

    @property
    def name(self):
        """The scientist's complete name."""
        return self._name

    @name.setter
    @accepts(str)
    def name(self, val):
        self._name = val

    @property
    def language(self):
        """The language(s) of the documents published until the given year."""
        return self._language

    @language.setter
    @accepts(str)
    def language(self, val):
        self._language = val

    @property
    def organization(self):
        """The affiliation name of the scientist's most frequent affiliation
        in the most recent year (before the given year) that
        the scientist published.
        """
        return self._organization

    @organization.setter
    @accepts(str)
    def organization(self, val):
        self._organization = val

    @property
    def publications(self):
        """The publications of the scientist published until
        the given year.
        """
        return self._publications

    @publications.setter
    @accepts((set, list, tuple))
    def publications(self, val):
        self._publications = val

    @property
    def publications_period(self):
        """The publications of the scientist published until
        the given year.
        """
        return self._publications_period

    @publications_period.setter
    @accepts((set, list, tuple))
    def publications_period(self, val):
        self._publications_period = val

    @property
    def sources(self):
        """The Scopus IDs of sources (journals, books) in which the
        scientist published until the given year.
        """
        return self._sources

    @sources.setter
    @accepts((list, tuple))
    def sources(self, val):
        self._sources = maybe_add_source_names(val, self.source_names)

    @property
    def surname(self):
        """The scientist's surname."""
        return self._surname

    @surname.setter
    @accepts(str)
    def surname(self, val):
        self._name = val

    @property
    def subjects(self):
        """The subject areas retrieved from Scopus AuthorSearch.
        """
        return self._subjects

    @subjects.setter
    @accepts((set, list, tuple))
    def subjects(self, val):
        self._subjects = val

    def __init__(self, identifier, year, refresh=False, period=None, eids=None):
        """Class to represent a scientist.

        Parameters
        ----------
        identifier : list of str
            List of Scopus Author IDs of the scientist.

        year : str or numeric
            Year for which characteristics should be defined for.

        refresh : boolean (optional, default=False)
            Whether to refresh all cached files or not.

        eids : list (optional, default=None)
            A list of scopus EIDs of the publications of the scientist.  If
            it is provided, the scientist's properties are set based on these
            publications, instead of the list of publications obtained from
            the Scopus Author ID(s).

        period: int (optional, default=None)
            The period in which to consider publications. If not provided,
            all publications are considered.

        Raises
        ------
        Exeption
            When there are no publications for the author until the
            given year.
        """
        self.identifier = identifier
        self.year = int(year)
        self.period = period
        self.year_period = None

        # Read mapping of fields to sources
        df, names = read_fields_sources_list()
        self.field_source = df
        self.source_names = names.set_index("source_id")["title"].to_dict()

        # Load list of publications
        if not eids:
            q = "AU-ID({})".format(") OR AU-ID(".join(identifier))
        else:
            q = "EID({})".format(" OR ".join(eids))
        res = base_query("docs", q, refresh,
            fields=["eid", "author_ids", "coverDate"])
        self._publications = [p for p in res if int(p.coverDate[:4]) <= year]
        if not len(self._publications):
            text = "No publications for author {} until year {}".format(
                "-".join(identifier), year)
            raise Exception(text)
        self._eids = eids or [p.eid for p in self._publications]

        # Fist year (if period provided set first year of period, if
        # not smaller than first year of publication
        pub_years = [p.coverDate[:4] for p in self._publications]
        self._first_year = int(min(pub_years))
        if period and year-period+1 <= self._first_year:
            self.period = None

        # Count of citations
        search_ids = eids or identifier
        self._citations = count_citations(search_ids=search_ids,
            pubyear=self.year+1, exclusion_key="AU-ID", exclusion_ids=identifier)

        # Coauthors
        self._coauthors = set(get_authors(self._publications)) - set(identifier)

        # Period counts simply set to total if period is or goes back to None
        if self.period:
            self.year_period = year-period+1
            pubs = [p for p in self._publications if
                    int(p.coverDate[:4]) <= year and
                    int(p.coverDate[:4]) >= self.year_period]
            self._publications_period = pubs
            if not len(self._publications_period):
                text = "No publications for author {} until year {} in a {}-"\
                       "years period".format("-".join(identifier), year,
                                             self.year_period)
                raise Exception(text)
            eids_period = [p.eid for p in self._publications_period]
            self._citations_period = count_citations(search_ids=eids_period,
                pubyear=self.year+1, exclusion_key="AU-ID",
                exclusion_ids=identifier)
            self._coauthors_period = set(get_authors(self._publications_period))
            self._coauthors_period -= set(identifier)
        else:
            self._coauthors_period = self._coauthors
            self._publications_period = self._publications
            self._citations_period = self._citations

        # Author search information
        source_ids = set([int(p.source_id) for p in self._publications
                          if p.source_id])
        self._sources = add_source_names(source_ids, self.source_names)
        self._active_year = int(max(pub_years))
        self._fields = df[df["source_id"].isin(source_ids)]["asjc"].tolist()
        self._main_field = get_main_field(self._fields)

        # Most recent geolocation
        ctry, afid, org = find_location(identifier, self._publications,
                                        year, refresh=refresh)
        self._country = ctry
        self._affiliation_id = afid
        self._organization = org
        self._language = None

        # Author name from profile with most documents
        au = query_author_data(self.identifier, refresh=refresh, verbose=False)
        au = au.sort_values("documents", ascending=False).iloc[0]
        self._subjects = [a.split(" ")[0] for a in au.areas.split("; ")]
        self._surname = au.surname or None
        if au.givenname:
            self._first_name = au.givenname.replace(".", " ").split(" ")[0]
        else:
            self._first_name = None
        if self._surname and au.givenname:
            self._name = ", ".join([self._surname, au.givenname])
        else:
            self._name = None

    def get_publication_languages(self, refresh=False):
        """Parse languages of published documents."""
        langs = []
        for eid in self._eids:
            l = AbstractRetrieval(eid, view="FULL", refresh=refresh).language
            langs.append(l)
        self._language = "; ".join(sorted(list(set(filter(None, langs)))))
        return self



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Tests for helpers module."""

import pandas as pd
from nose.tools import assert_equal, assert_true, raises

from sosia.classes import Scientist
from sosia.utils.helpers import add_source_names, flat_set_from_df,\
    margin_range, read_fields_sources_list


def test_add_source_names():
    s = Scientist(["55208373700"], 2017)
    expected = [(14351, "Brain Research Reviews"),
                (18632, "Progress in Brain Research")]
    ids, names = zip(*expected)
    received = add_source_names(ids, s.source_names)
    assert_equal(received, expected)


def test_flat_set_from_df():
    d = {'col1': [[1, 2], [10, 20]], "col2": ["a", "b"]}
    df = pd.DataFrame(d)
    expected = [1, 2, 10, 20]
    received = sorted(list(flat_set_from_df(df, "col1")))
    assert_equal(received, expected)


def test_flat_set_from_df_condition():
    d = {'col1': [[1, 2], [10, 20]], "col2": ["a", "b"]}
    df = pd.DataFrame(d)
    condition = df["col2"] == "b"
    expected = [10, 20]
    received = sorted(list(flat_set_from_df(df, "col1", condition)))
    assert_equal(received, expected)


def test_margin_range():
    assert_equal(margin_range(5, 1), range(4, 7))
    assert_equal(margin_range(10, 0.09), range(9, 12))


def test_read_fields_sources_list():
    sources, names = read_fields_sources_list()
    assert_equal(str(type(sources)), "<class 'pandas.core.frame.DataFrame'>")
    assert_equal(str(type(names)), "<class 'pandas.core.frame.DataFrame'>")



from nltk import snowball, word_tokenize


def clean_abstract(s):
    """Clean an abstract of a document."""
    # Remove copyright statement, which can be leading or trailing
    tokens = s.split(". ")
    if "©" in tokens[0]:
        return ". ".join(tokens[1:])
    for idx in (-2, -1):
        try:
            if "©" in tokens[idx]:
                return ". ".join(tokens[:idx]) + "."
        except:
            pass
    else:
        return s


def compute_cos(matrix, digits=4):
    """Compute cosine given a regular matrix."""
    return (matrix * matrix.T).toarray().round(digits)[-1][0]


def tokenize_and_stem(text):
    """Auxiliary function to return stemmed tokens of document"""
    return [snowball.SnowballStemmer("english").stem(t) for
            t in word_tokenize(text.lower())]



from os.path import expanduser

URL_CONTENT = "https://www.elsevier.com/solutions/scopus/how-scopus-works/content"
URL_SOURCES = "https://elsevier.com/?a=734751"

FIELDS_SOURCES_LIST = expanduser("~/.sosia/") + "field_sources_list.csv"
SOURCES_NAMES_LIST = expanduser("~/.sosia/") + "sources_names.csv"
CONFIG_FILE = expanduser("~/.sosia/config.ini")

ASJC_2D = {10: "MULT", 11: "AGRI", 12: "ARTS", 13: "BIOC", 14: "BUSI",
           15: "CENG", 16: "CHEM", 17: "COMP", 18: "DECI", 19: "EART",
           20: "ECON", 21: "ENER", 22: "ENGI", 23: "ENVI", 24: "IMMU",
           25: "MATE", 26: "MATH", 27: "MEDI", 28: "NEUR", 29: "NURS",
           30: "PHAR", 31: "PHYS", 32: "PSYC", 33: "SOCI", 34: "VETE",
           35: "DENT", 36: "HEAL"}

CACHE_TABLES = {
    "author_cits_size":
        {"columns": (("auth_id", "int"), ("year", "int"), ("n_cits", "int")),
         "primary": ("auth_id", "year")},
    "author_size":
        {"columns": (("auth_id", "int"), ("year", "int"), ("n_pubs", "int")),
         "primary": ("auth_id", "year")},
    "author_year":
        {"columns": (("auth_id", "int"), ("year", "int"), ("first_year", "int"),
                     ("n_pubs", "int"), ("n_coauth", "int")),
         "primary": ("auth_id", "year")},
    "authors":
        {"columns": (("auth_id", "int"), ("eid", "text"), ("surname", "text"),
                     ("initials", "text"), ("givenname", "text"),
                     ("affiliation", "text"), ("documents", "text"),
                     ("affiliation_id", "text"), ("city", "text"),
                     ("country", "text"), ("areas", "text")),
         "primary": ("auth_id",)},
    "sources":
        {"columns": (("source_id", "int"), ("year", "int"), ("auids", "text")),
         "primary": ("source_id", "year")},
    "sources_afids":
        {"columns": (("source_id", "int"), ("year", "int"), ("afid", "int"),
                     ("auids", "text")),
         "primary": ("source_id", "year", "afid")}}



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Tests for queries module."""

from nose.tools import assert_equal, assert_true
from string import Template
import pandas as pd

from sosia.processing import base_query, count_citations, query_author_data,\
    query_journal, query_year, stacked_query


def test_base_query():
    auth_id = 53164702100
    q = "AU-ID({}) AND PUBYEAR BEF {}".format(auth_id, 2017)
    size = base_query("docs", q, size_only=True)
    assert_equal(size, 5)


def test_base_query_author():
    auth_id = 53164702100
    query = "AU-ID({})".format(auth_id)
    size = base_query("author", query, size_only=True)
    assert_equal(size, 1)


def test_count_citations():
    identifier = ["55208373700", "55208373700"]
    count1 = count_citations(identifier, 2017, "AU-ID", exclusion_ids=identifier)
    assert_equal(count1, 22)
    eids = ["2-s2.0-84959420483", "2-s2.0-84949113230"]
    count2 = count_citations(eids, 2017, "AU-ID", exclusion_ids=identifier)
    assert_equal(count2, 1)
    eids_long = eids * 200
    count3 = count_citations(eids_long, 2017, "AU-ID", exclusion_ids=identifier)
    assert_equal(count3, 4)


def test_query_author_data():
    auth_list = [6701809842, 55208373700]
    auth_data = query_author_data(auth_list, refresh=False, verbose=False)
    assert_true(isinstance(auth_data,  pd.DataFrame))
    expected_cols = ["auth_id", "eid", "surname", "initials", "givenname",
                     "affiliation", "documents", "affiliation_id", "city",
                     "country", "areas"]
    assert_equal(auth_data.columns.tolist(), expected_cols)
    assert_equal(auth_data.auth_id.tolist(), auth_list)
    assert_equal(auth_data.surname.tolist(), ["Harhoff", "Baruffaldi"])


def test_query_journal():
    # test a journal with more than 5k publications in one year
    res = query_journal("11000153773", [2006], refresh=False)
    assert_true(24100 < len(res.get("2006")) < 25000)


def test_query_year():
    # test a journal and year
    res = query_year(2010, [22900], refresh=False, verbose=False)
    assert_equal(res["source_id"].tolist(), ['22900'])
    assert_equal(res["year"].tolist(), ['2010'])
    assert_true(isinstance(res["auids"][0], list))
    assert_true(len(res["auids"][0]) > 0)
    # test a journal and year that are not in scopus
    res = query_year(1969, [22900], refresh=False, verbose=False)
    assert_true(res.empty)
    # test a large query (>5000 scopus results)
    source_ids = [13703, 13847, 13945, 14131, 14150, 14156, 14204, 14207,
                  14209, 14346, 14438, 14536, 14539, 15034, 15448, 15510, 15754]
    res = query_year(1984, source_ids, refresh=False, verbose=True)
    assert_true(len(res[~res["auids"].isnull()]) == 17)
    assert_true(isinstance(res["auids"][0], list))
    assert_true(len(res["auids"][0]) > 0)


def test_query_year_afid():
    # test keeping affiliation id column
    source_ids = [13703, 13847, 13945, 14131, 14150, 14156, 14204, 14207,
                  14209, 14346, 14438, 14536, 14539, 15034, 15448, 15510, 15754]
    res = query_year(1984, source_ids, refresh=False, verbose=True, afid=True)
    expected = range(3077-5, 3077+5)
    assert_true(len(res[~res["auids"].isnull()]) in expected)
    assert_true(res.columns.tolist(), ['source_id', 'year', 'afid', 'auids'])
    assert_true(len(res["auids"][0]) > 0)


def test_stacked_query():
    # test a query with one journal-year above 5000
    group = [18400156716, 19300157101, 19400157208, 19400157312, 19500157223,
             19600166213, 19700175482, 19700182353, 19800188009, 19900193211,
             20100195028, 21100208103, 21100225839, 21100228010, 21100244622,
             21100246535, 21100246537, 21100285035, 21100313905, 21100329904,
             21100370441, 21100370876, 21100416121, 21100775937, 21100871308,
             25674]
    template = Template("SOURCE-ID($fill) AND PUBYEAR IS {}".format(1998))
    joiner = " OR "
    refresh = False
    q_type = "docs"
    res = []
    stacked_query(group, res, template, joiner, q_type, refresh)
    assert_equal(len(res), 6441)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Tests for extraction module."""

from nose.tools import assert_equal, assert_true, assert_false
from pybliometrics.scopus import ScopusSearch
import pandas as pd

from sosia.processing import expand_affiliation, find_location, parse_docs


def test_expand_affiliation():
    auth_id = 6701809842
    pubs = ScopusSearch("AU-ID({})".format(auth_id)).results
    res = pd.DataFrame(pubs)
    res = expand_affiliation(res)
    assert_equal(len(res), 185)
    expect_columns = ['source_id', 'author_ids', 'afid']
    assert_equal(set(res.columns.tolist()), set(expect_columns))
    assert_true(any(res.author_ids.str.contains(";")))
    assert_false(any(res.afid.str.contains(";")))


def test_find_location():
    auth_id = 6701809842
    pubs = ScopusSearch("AU-ID({})".format(auth_id)).results
    ctry, aid, aff = find_location([str(auth_id)], pubs, 2000, refresh=False)
    assert_equal(ctry, "Germany")
    assert_equal(aid, "60028717")
    assert_equal(aff, "University of Munich")


def test_parse_docs():
    eids = ["2-s2.0-84866317084"]
    received = parse_docs(eids, refresh=False)
    expected_refs = ('29144517611 57849112238 51249091642 70449099678 '
                     '84865231386 15944370019 8744256776 0004256525 '
                     '84866333650 78650692566 0002969912 0007622058 '
                     '0000169440 0003685848 43049125937 43149086011 '
                     '84866332328 27744606594 2442709303 84866309814 '
                     '34248571923 0029824384 34548317343 3142544611 '
                     '84981198417 0003457562 0032394091 0001116544 '
                     '84866324358 23944478214 0039484749 0001275239 '
                     '34249696486 70449641263 0035654659 0041099301 '
                     '55049124635 67650248718 39149084479 40749109418 '
                     '35348862941 0030559826 34547860480 77953702055 '
                     '18144372897 0004062815 84972591424 0034423919 '
                     '0033675552 0034424078 0027767147 0035654590 '
                     '4243112264 29144486677 0004164119 85015576166 '
                     '33645383647 84981213068 23044470851 78649697033 '
                     '2442654430 0141625872 30444461409 0034435025 '
                     '47949124687 84920182751 84887864855 84866332329 '
                     '84984932935 33845620645 0942299814')
    assert_equal(received[0], expected_refs)
    expected_abs = (
        "Through an analysis of 497 foreign researchers in Italy "
        "and Portugal we verify the impact of home linkages on return "
        "mobility choices and scientific productivity. We consider the "
        "presence of several different types of linkages of the researchers "
        "working abroad with their country of origin and control for the "
        "most relevant contextual factors (age, research area, position in "
        "the host country, etc.). The probability of return to their home "
        "country and scientific productivity in the host country are both "
        "higher for researchers that maintain home linkages. We conclude "
        "that the presence of home linkages directly benefits both "
        "countries in addition to the indirect benefit of expanding the "
        "scientific networks. Policy implications and suggestions for "
        "further research are discussed."
    )
    assert_equal(received[2], expected_abs)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Tests for startup module."""


from os import remove
from os.path import expanduser, isfile

import pandas as pd
from nose.tools import assert_equal, assert_true

from sosia.utils.constants import FIELDS_SOURCES_LIST, SOURCES_NAMES_LIST
from sosia.utils.startup import create_fields_sources_list, create_cache

# try:
#     remove(FIELDS_SOURCES_LIST)
# except FileNotFoundError:
#     pass
# try:
#     remove(SOURCES_NAMES_LIST)
# except FileNotFoundError:
#     pass
# create_fields_sources_list()


def test_create_cache():
    cache_file = expanduser("~/.sosia/") + "cache_sqlite.sqlite"
    cache_file_test = expanduser("~/.sosia/") + "cache_sqlite_test.sqlite"
    create_cache()
    assert_true(isfile(cache_file))
    create_cache(file=cache_file_test)
    assert_true(isfile(cache_file_test))


def test_fields_sources_list():
    df = pd.read_csv(FIELDS_SOURCES_LIST)
    assert_true(isinstance(df, pd.DataFrame))
    assert_equal(list(df.columns), ["asjc", "source_id", "type"])
    assert_true(df.shape[0] > 55130)


def test_sources_names_list():
    df = pd.read_csv(SOURCES_NAMES_LIST, index_col=0)
    assert_true(isinstance(df, pd.DataFrame))
    assert_equal(list(df.columns), ["title"])
    assert_true(df.shape[0] > 74000)
    assert_equal(df.loc[12005].title, "Journal of Traumatic Stress")
    assert_equal(df.loc[21100875102].title, "Diagnostyka")



from copy import deepcopy

import numpy as np
import pandas as pd
import sqlite3

from sosia.utils import config, flat_set_from_df

cache_file = config.get('Cache', 'File path')


def cache_connect(file=cache_file):
    """Connect to cache file.

    Parameters
    ----------
    file : file (optional, default=cache_file)
        The cache file to connect to.
    """
    file = config.get('Cache', 'File path')
    conn = sqlite3.connect(file)
    c = conn.cursor()
    sqlite3.register_adapter(np.int64, lambda val: int(val))
    sqlite3.register_adapter(np.int32, lambda val: int(val))
    return c, conn


def cache_insert(data, table, file=cache_file):
    """Insert new authors information in cache.

    Parameters
    ----------
    data : DataFrame or 3-tuple
        Dataframe with authors information or (when table="source") a
        3-element tuple.

    table : string
        The database table to insert into.  The query will be adjusted
        accordingly.
        Allowed values: "authors", "author_cits_size", "author_year",
        "author_size", "sources".

    file : file (optional, default=cache_file)
        The cache file to connect to.

    Raises
    ------
    ValueError
        If parameter table is not one of the allowed values.
    """
    def join_flat_auids(s):
        return ",".join([str(a) for a in s["auids"]])

    # Build query
    if table == 'authors':
        q = """INSERT OR IGNORE INTO authors (auth_id, eid, surname, initials,
            givenname, affiliation, documents, affiliation_id, city, country,
            areas) values (?,?,?,?,?,?,?,?,?,?,?)"""
        if data.empty:
            return None
        data["auth_id"] = data.apply(lambda x: x.eid.split("-")[-1], axis=1)
        cols = ["auth_id", "eid", "surname", "initials", "givenname",
                "affiliation", "documents", "affiliation_id", "city",
                "country", "areas"]
        data = data[cols]
    elif table == 'author_cits_size':
        q = """INSERT OR IGNORE INTO author_cits_size (auth_id, year, n_cits)
            values (?,?,?)"""
    elif table == 'author_year':
        q = """INSERT OR IGNORE INTO author_year (auth_id, year, first_year, n_pubs,
            n_coauth) values (?,?,?,?,?) """
    elif table == 'author_size':
        q = """INSERT OR IGNORE INTO author_size (auth_id, year, n_pubs)
            values ({},{},{}) """.format(data[0], data[1], data[2])
    elif table == "sources":
        if data.empty:
            return None
        if "afid" in data.columns.tolist():
            data = (data.groupby(["source_id", "year"])[["auids"]]
                    .apply(lambda x: list(flat_set_from_df(x, "auids")))
                    .rename("auids")
                    .reset_index())
        data["auids"] = data.apply(join_flat_auids, axis=1)
        data = data[["source_id", "year", "auids"]]
        q = """INSERT OR IGNORE INTO sources (source_id, year, auids)
            VALUES (?,?,?) """
    elif table == "sources_afids":
        if data.empty:
            return None
        data["auids"] = data.apply(join_flat_auids, axis=1)
        data = data[["source_id", "year", "afid", "auids"]]
        q = """INSERT OR IGNORE INTO sources_afids (source_id, year, afid, auids)
            VALUES (?,?,?,?) """
    else:
        allowed_tables = ('authors', 'author_cits_size', 'author_year',
                          'author_size', 'sources', 'sources_afids')
        msg = 'table parameter must be one of ' + ', '.join(allowed_tables)
        raise ValueError(msg)
    # Perform caching
    _, conn = cache_connect(file=file)
    if table in ("authors", "author_cits_size", "author_year", "sources",
                 "sources_afids"):
        conn.executemany(q, data.to_records(index=False))
    else:
        conn.execute(q)
    conn.commit()


def insert_temporary_table(df, merge_cols, file=cache_file):
    """Temporarily create a table in SQL cache in order to prepare a
    merge with `table`.

    Parameters
    ----------
    data : DataFrame
        Dataframe with authors information that should be entered.

    merge_cols : list of str
        The columns that should be created and filled.  Must correspond in
        length to the number of columns of `df`.

    file : file (optional, default=cache_file)
        The cache file to connect to.
    """
    df = df.astype({c: "int64" for c in merge_cols})
    c, conn = cache_connect(file=file)
    # Drop table
    c.execute("DROP TABLE IF EXISTS temp")
    # Create table
    names = ", ".join(merge_cols)
    q = "CREATE TABLE temp ({0}, PRIMARY KEY({0}))".format(names)
    c.execute(q)
    # Insert values
    wildcards = ", ".join(["?"] * len(merge_cols))
    q = "INSERT OR IGNORE INTO temp ({}) values ({})".format(names, wildcards)
    conn.executemany(q, df.to_records(index=False))
    conn.commit()


def d_to_df_for_cache(d, source_id):
    """Function to create a DataFrame of sources, years and list of authors
    from a dictionary where keys are the years and values are the list of
    authors.

    Parameters
    ----------
    d : dict
        Dictionary of year-list of author pairs.

    source_id: int
        Scopus identifier of the source.
    """
    d2 = deepcopy(d)
    for y in d2:
        d2[y] = [d2[y]]
    df = pd.DataFrame.from_dict(d2, orient="index").reset_index()
    df.columns = ["year", "auids"]
    df["source_id"] = str(source_id)
    return df



from collections import namedtuple
from operator import attrgetter
import pandas as pd

from pybliometrics.scopus import AbstractRetrieval
from pybliometrics.scopus.exception import Scopus404Error
from sklearn.feature_extraction.text import TfidfVectorizer

from sosia.processing.nlp import clean_abstract, compute_cos, tokenize_and_stem
from sosia.utils import print_progress


def expand_affiliation(df):
    """Auxiliary function to expand the information about the affiliation
    in publications from ScopusSearch.
    """
    res = df[["source_id", "author_ids", "afid"]].copy()
    res['afid'] = res["afid"].str.split(';')
    res = (res["afid"].apply(pd.Series)
              .merge(res, right_index=True, left_index=True)
              .drop(["afid"], axis=1)
              .melt(id_vars=['source_id', 'author_ids'], value_name="afid")
              .drop("variable", axis=1)
              .dropna())
    return res


def find_location(auth_ids, pubs, year, refresh):
    """Find the most common country, affiliation ID, and affiliation name
    of a scientist using her most recent publications with valid information.

    Parameters
    ----------
    auth_ids : list of str
        A list of Scopus Author Profile IDs for which the affiliation should
        be searched for.

    pubs : list of namedtuple
        The publications associated with the Author IDs as returned from a
        scopus query.

    year : int
        The year for which we would like to have the country.

    refresh : bool
        Whether to refresh all cached files or not.

    Returns
    -------
    country, affiliation_id, organization : str or None
        The country, city, affiliation ID, and affiliation name of the
        scientist in the year closest to the given year, given that the
        publications list valid information for each output. Equals None when
        no valid publications are found.
    """
    # Available papers of most recent year with publications
    papers = [p for p in pubs if int(p.coverDate[:4]) <= year]
    papers = sorted(papers, key=attrgetter("coverDate"), reverse=True)
    params = {"view": "FULL", "refresh": refresh}
    # Return most recent complete information
    for p in papers:
        try:
            authgroup = AbstractRetrieval(p.eid, **params).authorgroup or []
        except Scopus404Error:
            continue
        authgroup = [a for a in authgroup if a.auid in auth_ids
                     and a.country and a.affiliation_id and a.organization]
        countries = "; ".join(sorted(set([a.country for a in authgroup])))
        aff_ids = "; ".join(sorted(set([a.affiliation_id for a in authgroup])))
        orgs = "; ".join(sorted(set([a.organization for a in authgroup])))
        if not countries and not aff_ids and not orgs:
            continue
        return (countries, aff_ids, orgs)
    # Return None-triple if all else fails
    return (countries, aff_ids, orgs)


def get_authors(pubs):
    """Get list of author IDs from a list of namedtuples representing
    publications.
    """
    l = [x.author_ids.split(";") for x in pubs if isinstance(x.author_ids, str)]
    return [au for sl in l for au in sl]


def get_auth_from_df(pubs):
    """Get list of author IDs from a dataframe of publications.
    """
    l = [x.split(";") for x in pubs.author_ids if isinstance(x, str)]
    return [au for sl in l for au in sl]


def inform_matches(profiles, focal, keywords, stop_words, verbose,
                   refresh, **kwds):
    """Create namedtuple adding information to matches.

    Parameters
    ----------
    profiles : list of Scientist()
        A list of Scientist objects representing matches.

    focal : Scientist
        Object of class Scientist representing the focal scientist.

    keywords : iterable of strings
        Which information to add to matches.

    stop_words : list
        A list of words that should be filtered in the analysis of abstracts.

    verbose : bool
        Whether to report on the progress of the process and the completeness
        of document information.

    refresh : bool
        Whether to refresh all cached files or not.

    kwds : keywords
        Parameters to pass to sklearn.feature_extraction.text.TfidfVectorizer
        for abstract and reference vectorization.

    Returns
    -------
    m : list of namedtuples
        A list of namedtuples representing matches.  Provided information
        depend on provided keywords.
    """
    from sosia.classes import Scientist
    # Create Match object
    fields = "ID name " + " ".join(keywords)
    m = namedtuple("Match", fields)
    # Preparation
    doc_parse = "reference_sim" in keywords or "abstract_sim" in keywords
    total = len(profiles)
    print_progress(0, total, verbose)
    if doc_parse:
        focal_eids = [d.eid for d in focal.publications]
        focal = parse_docs(focal_eids, refresh)
        focal_refs, focal_refs_n, focal_abs, focal_abs_n = focal
    # Add selective information
    out = []
    info = {}  # to collect information on missing information
    for idx, p in enumerate(profiles):
        # Add characteristics
        match_info = {"ID": p.identifier[0], "name": p.name}
        if "language" in keywords:
            try:
                match_info["language"] = p.get_publication_languages().language
            except Scopus404Error:  # Refresh profile
                p = Scientist(p.identifier, p.year, refresh=True)
                match_info["language"] = p.get_publication_languages().language
        if "first_name" in keywords:
            match_info["first_name"] = p.first_name
        if "surname" in keywords:
            match_info["surname"] = p.surname
        if "first_year" in keywords:
            match_info["first_year"] = p.first_year
        if "num_coauthors" in keywords:
            match_info["num_coauthors"] = len(p.coauthors)
        if "num_publications" in keywords:
            match_info["num_publications"] = len(p.publications)
        if "num_citations" in keywords:
            match_info["num_citations"] = p.citations
        if "num_coauthors_period" in keywords:
            match_info["num_coauthors_period"] = len(p.coauthors_period)
        if "num_publications_period" in keywords:
            match_info["num_publications_period"] = len(p.publications_period)
        if "num_citations_period" in keywords:
            match_info["num_citations_period"] = p.citations_period
        if "subjects" in keywords:
            match_info["subjects"] = p.subjects
        if "country" in keywords:
            match_info["country"] = p.country
        if "city" in keywords:
            match_info["city"] = p.city
        if "affiliation_id" in keywords:
            match_info["affiliation_id"] = p.affiliation_id
        if "affiliation" in keywords:
            match_info["affiliation"] = p.organization
        # Abstract and reference similiarity is performed jointly
        if doc_parse:
            eids = [d.eid for d in p.publications]
            refs, refs_n, absts, absts_n = parse_docs(eids, refresh)
            vec = TfidfVectorizer(**kwds)
            ref_cos = compute_cos(vec.fit_transform([refs, focal_refs]))
            vec = TfidfVectorizer(stop_words=stop_words,
                                  tokenizer=tokenize_and_stem, **kwds)
            abs_cos = compute_cos(vec.fit_transform([absts, focal_abs]))
            # Save info for below print statement
            meta = namedtuple("Meta", "refs absts total")
            meta(refs=refs_n, absts=absts_n, total=len(eids))
            key = "; ".join(p.identifier)
            info[key] = meta(refs=refs_n, absts=absts_n, total=len(eids))
        if "reference_sim" in keywords:
            match_info["reference_sim"] = ref_cos
        if "abstract_sim" in keywords:
            match_info["abstract_sim"] = abs_cos
        # Finalize
        out.append(m(**match_info))
        print_progress(idx+1, total, verbose)
    # Print information on missing information
    if verbose and doc_parse:
        for auth_id, info in info.items():
            _print_missing_docs(auth_id, info.refs, info.absts, info.total)
        label = ";".join(focal.identifier) + " (focal)"
        focal_pubs_n = len(focal.publications)
        _print_missing_docs(label, focal_refs_n, focal_abs_n, focal_pubs_n)
    return out


def parse_docs(eids, refresh):
    """Find abstract and references of articles published up until
    the given year, both as continuous string.

    Parameters
    ----------
    eids : list of str
        Scopus Document EIDs representing documents to be considered.

    refresh : bool
        Whether to refresh the cached files if they exist, or not.

    Returns
    -------
    t : tuple
        A tuple with our elements: The first element is a continuous string
        of cleaned abstracts, joined on a blank.  The second element is the
        number of documents with valid reference information.  The third
        element is a continuous string of Scopus Abstract EIDs representing
        cited references, joined on a blank.  The fourth element is the
        number of valid abstract information.
    """
    docs = []
    for eid in eids:
        try:
            docs.append(AbstractRetrieval(eid, view="FULL", refresh=refresh))
        except Scopus404Error:
            continue
    refs = [ab.references for ab in docs if ab.references]
    valid_refs = len(refs)
    refs = " ".join([ref.id for sl in refs for ref in sl])
    absts = [clean_abstract(ab.abstract) for ab in docs if ab.abstract]
    valid_absts = len(absts)
    absts = " ".join(absts)
    return (refs, valid_refs, absts, valid_absts)


def _print_missing_docs(auth_id, valid_abs, valid_refs, total, verbose=True):
    """Auxiliary function to print information on missing abstracts and
    reference lists stored in a dictionary d.
    """
    miss_abs = total-valid_abs
    miss_refs = total-valid_refs
    print("Researcher {}: {} abstract(s) and {} reference list(s) out of "
          "{} documents missing".format(auth_id, miss_abs, miss_refs, total))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Tests for cache module."""

from nose.tools import assert_equal, assert_true
from itertools import product
from os.path import expanduser

import pandas as pd
from pybliometrics.scopus import ScopusSearch, AuthorSearch
from pandas.testing import assert_frame_equal
from sosia.utils import config

from sosia.cache import authors_in_cache, author_size_in_cache,\
    author_year_in_cache, cache_insert, sources_in_cache
from sosia.processing import query_year
from sosia.utils import build_dict, create_cache


test_cache = expanduser("~/.sosia/") + "cache_sqlite_test.sqlite"
config["Cache"]["File path"] = test_cache

def test_authors_in_cache():
    create_cache(drop=True, file=test_cache)
    # Variables
    expected_auth = ["53164702100", "57197093438"]
    search_auth = ["55317901900"]
    # Test empty cache
    df1 = pd.DataFrame(expected_auth, columns=["auth_id"], dtype="int64")
    incache, tosearch = authors_in_cache(df1, file=test_cache)
    expected_cols = ['auth_id', 'eid', 'surname', 'initials', 'givenname',
                     'affiliation', 'documents', 'affiliation_id', 'city',
                     'country', 'areas']
    expected_auth = [int(au) for au in expected_auth]
    assert_equal(tosearch, expected_auth)
    assert_equal(len(incache), 0)
    assert_equal(incache.columns.tolist(), expected_cols)
    # Test partial retrieval
    q = "AU-ID({})".format(') OR AU-ID('.join([str(a) for a in expected_auth]))
    res = pd.DataFrame(AuthorSearch(q).authors, dtype="int64")
    res["auth_id"] = res["eid"].str.split("-").str[-1]
    res = res[expected_cols]
    cache_insert(res, table="authors", file=test_cache)
    df2 = pd.DataFrame(expected_auth + search_auth, columns=["auth_id"],
                       dtype="int64")
    incache, tosearch = authors_in_cache(df2, file=test_cache)
    assert_equal(tosearch, [55317901900])
    assert_equal(len(incache), 2)
    # Test full retrieval
    incache, tosearch = authors_in_cache(df1, file=test_cache)
    assert_equal(tosearch, [])
    assert_equal(len(incache), 2)


def test_author_year_in_cache():
    create_cache(drop=True, file=test_cache)
    # Variables
    expected_auth = ["53164702100", "57197093438"]
    search_auth = ["55317901900"]
    year = 2016
    # Test empty cache
    df1 = pd.DataFrame(expected_auth, columns=["auth_id"],
                       dtype="int64")
    df1["year"] = year
    auth_y_incache, auth_y_search = author_year_in_cache(df1, file=test_cache)
    assert_frame_equal(auth_y_search, df1)
    assert_equal(len(auth_y_incache), 0)
    # Test partial retrieval
    fill = ') OR AU-ID('.join([str(a) for a in expected_auth])
    q = "(AU-ID({})) AND PUBYEAR BEF {}".format(fill, year+1)
    res = build_dict(ScopusSearch(q).results, expected_auth)
    res = pd.DataFrame.from_dict(res, orient="index", dtype="int64")
    res["year"] = year
    cols = ["year", "first_year", "n_pubs", "n_coauth"]
    res = res[cols].reset_index().rename(columns={"index": "auth_id"})
    cache_insert(res, table="author_year", file=test_cache)
    df2 = pd.DataFrame(expected_auth + search_auth,
                       columns=["auth_id"], dtype="int64")
    df2["year"] = year
    auth_y_incache, auth_y_search = author_year_in_cache(df2, file=test_cache)
    expected_auth = [int(au) for au in expected_auth]
    search_auth = [int(au) for au in search_auth]
    assert_equal(sorted(auth_y_incache.auth_id.tolist()), expected_auth)
    assert_equal(auth_y_incache.year.tolist(), [year, year])
    assert_equal(auth_y_search.auth_id.tolist(), search_auth)
    assert_equal(auth_y_search.year.tolist(), [year])
    # Test full retrieval
    auth_year_incache, auth_year_search = author_year_in_cache(df1, file=test_cache)
    assert_equal(sorted(auth_year_incache.auth_id.tolist()), expected_auth)
    assert_equal(auth_year_incache.year.tolist(), [year, year])
    assert_true(auth_year_search.empty)


def test_author_size_in_cache():
    create_cache(drop=True, file=test_cache)
    # Variables
    expected_auth = 53164702100
    expected_years = [2010, 2017]
    pubs1 = 0
    pubs2 = 6
    cols = ["auth_id", "year"]
    df = pd.DataFrame(list(product([expected_auth], expected_years)),
                      columns=cols, dtype="int64")
    # Test empty cache
    size = author_size_in_cache(df, file=test_cache)
    assert_equal(len(size), 0)
    assert_true(isinstance(size, pd.DataFrame))
    # Test adding to and retrieving from cache
    tp1 = (expected_auth, expected_years[0], pubs1)
    cache_insert(tp1, table="author_size", file=test_cache)
    tp2 = (expected_auth, expected_years[1], pubs2)
    cache_insert(tp2, table="author_size", file=test_cache)
    size = author_size_in_cache(df, file=test_cache)
    assert_equal(len(size), 2)
    assert_frame_equal(size[cols], df)
    assert_equal(size[size.year == expected_years[0]]["n_pubs"][0], pubs1)
    assert_equal(size[size.year == expected_years[1]]["n_pubs"][1], pubs2)


def test_sources_in_cache_empty():
    create_cache(drop=True, file=test_cache)
    df = pd.DataFrame(list(product([22900], [2010, 2005])),
                      columns=["source_id", "year"], dtype="int64")
    sources_ys_incache, sources_ys_search = sources_in_cache(df, file=test_cache)
    assert_frame_equal(sources_ys_search, df)
    assert_true(sources_ys_incache.empty)


def test_sources_in_cache_partial():
    create_cache(drop=True, file=test_cache)
    # Variables
    expected_sources = [22900]
    expected_years = [2010, 2005]
    df = pd.DataFrame(list(product(expected_sources, expected_years)),
                      columns=["source_id", "year"], dtype="int64")
    # Populate cache
    res = query_year(expected_years[0], expected_sources, False, False)
    cache_insert(res, table="sources", file=test_cache)
    # Retrieve from cache
    sources_ys_incache, sources_ys_search = sources_in_cache(df, file=test_cache)
    expected_sources = [int(s) for s in expected_sources]
    assert_equal(sources_ys_incache.source_id.tolist(), expected_sources)
    assert_equal(sources_ys_incache.year.tolist(), [expected_years[0]])
    assert_equal(sources_ys_search.source_id.tolist(), expected_sources)
    assert_equal(sources_ys_search.year.tolist(), [expected_years[1]])


def test_sources_in_cache_full():
    create_cache(drop=True, file=test_cache)
    # Variables
    expected_sources = [22900]
    expected_years = [2010, 2005]
    cols = ["source_id", "year"]
    df = pd.DataFrame(list(product(expected_sources, expected_years)),
                      columns=cols, dtype="int64")
    # Populate cache
    res = query_year(expected_years[0], expected_sources, False, False)
    cache_insert(res, table="sources", file=test_cache)
    sources_ys_incache, sources_ys_search = sources_in_cache(df, file=test_cache)
    # Retrieve from cache
    sources_ys = sources_ys_incache[cols]
    sources_ys_incache, sources_ys_search = sources_in_cache(sources_ys,
                                                             file=test_cache)
    expected_sources = [int(s) for s in expected_sources]
    assert_equal(sources_ys_incache.source_id.tolist(), expected_sources)
    assert_equal(sources_ys_incache.year.tolist(), [expected_years[0]])
    assert_true(sources_ys_search.empty)
    

def test_sources_afids_in_cache_empty():
    create_cache(drop=True, file=test_cache)
    df = pd.DataFrame(list(product([22900], [2010, 2005])),
                      columns=["source_id", "year"], dtype="int64")
    sa_incache, sa_search = sources_in_cache(df, file=test_cache, afid=True)
    assert_frame_equal(sa_search, df)
    assert_true(sa_incache.empty)


def test_sources_afids_in_cache_partial():
    create_cache(drop=True, file=test_cache)
    # Variables
    expected_sources = [22900]
    expected_years = [2010, 2005]
    df = pd.DataFrame(list(product(expected_sources, expected_years)),
                      columns=["source_id", "year"], dtype="int64")
    sa_incache, sa_search = sources_in_cache(df, file=test_cache, afid=True)
    # Populate cache
    res = query_year(expected_years[0], expected_sources, False, False, afid=True)
    cache_insert(res, table="sources_afids", file=test_cache)
    # Retrieve from cache
    sa_incache, sa_search = sources_in_cache(df, file=test_cache, afid=True)
    expected_sources = set([int(s) for s in expected_sources])
    assert_equal(set(sa_incache.source_id.tolist()), set(expected_sources))
    assert_equal(set(sa_incache.year.tolist()), set([expected_years[0]]))
    assert_equal(set(sa_search.source_id.tolist()), set(expected_sources))
    assert_equal(set(sa_search.year.tolist()), set([expected_years[1]]))
    expected = range(182-5, 182+5)
    assert_true(len(sa_incache) in expected)
    assert_true(len(sa_incache.afid.drop_duplicates()) in expected)


def test_sources_afids_in_cache_full():
    create_cache(drop=True, file=test_cache)
    # Variables
    expected_sources = [22900]
    expected_years = [2010, 2005]
    cols = ["source_id", "year"]
    df = pd.DataFrame(list(product(expected_sources, expected_years)),
                      columns=cols, dtype="int64")
    # Populate cache
    res = query_year(expected_years[0], expected_sources, False, False, afid=True)
    cache_insert(res, table="sources_afids", file=test_cache)
    # Retrieve from cache
    sa_incache, sa_search = sources_in_cache(df, file=test_cache, afid=True)
    sources_ys = sa_incache[cols].drop_duplicates()
    sa_incache, sa_search = sources_in_cache(sources_ys, file=test_cache,
                                             afid=True)
    expected_sources = set([int(s) for s in expected_sources])
    assert_equal(set(sa_incache.source_id.tolist()), set(expected_sources))
    assert_equal(set(sa_incache.year.tolist()), set([expected_years[0]]))
    expected = range(182-5, 182+5)
    assert_true(len(sa_incache) in expected)
    assert_true(len(sa_incache.afid.drop_duplicates()) in expected)
    assert_true(sa_search.empty)


def test_sources_afids_in_sources_cache():
    create_cache(drop=True, file=test_cache)
    # Variables
    expected_sources = [22900]
    expected_years = [2010, 2005]
    df = pd.DataFrame(list(product(expected_sources, expected_years)),
                      columns=["source_id", "year"], dtype="int64")
    # Populate cache
    res = query_year(expected_years[0], expected_sources, False, False, afid=True)
    cache_insert(res, table="sources", file=test_cache)
    # Retrieve from cache
    sources_ys_incache, sources_ys_search = sources_in_cache(df, file=test_cache)
    expected_sources = [int(s) for s in expected_sources]
    assert_equal(sources_ys_incache.source_id.tolist(), expected_sources)
    assert_equal(sources_ys_incache.year.tolist(), [expected_years[0]])
    assert_equal(sources_ys_search.source_id.tolist(), expected_sources)
    assert_equal(sources_ys_search.year.tolist(), [expected_years[1]])




#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Tests for nlp module."""

from string import digits, punctuation
import warnings

from nose.tools import assert_equal, assert_true
from numpy import array
from scipy.sparse import csr_matrix
from sklearn.feature_extraction.stop_words import ENGLISH_STOP_WORDS

from sosia.processing import clean_abstract, compute_cos, tokenize_and_stem

warnings.filterwarnings("ignore")
STOPWORDS = list(ENGLISH_STOP_WORDS)
STOPWORDS.extend(punctuation + digits)


def test_clean_abstract():
    expected = "Lorem ipsum."
    assert_equal(clean_abstract("Lorem ipsum. © dolor sit."), expected)
    assert_equal(clean_abstract("© dolor sit. Lorem ipsum."), expected)
    assert_equal(clean_abstract(expected), expected)


def test_compute_cos():
    expected = 0.6875
    received = compute_cos(csr_matrix(array([[0.5, 0.75], [1, 0.25]])))
    assert_equal(received, expected)


def test_tokenize_and_stem():
    expected = ["lorem", "1", "2", "3", "ipsum"]
    received = tokenize_and_stem("Lorem 1 2 3 Ipsum")
    assert_equal(received, expected)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Tests for class `Scientist`."""

from collections import namedtuple
from nose.tools import assert_equal, assert_true, assert_false

from sosia.classes import Scientist

scientist1 = Scientist(["6701809842"], 2001, refresh=True)
scientist2 = Scientist(["55208373700", "55208373700"], 2017, refresh=True)
eids = ["2-s2.0-84959420483", "2-s2.0-84949113230"]
scientist3 = Scientist(["55208373700"], 2017, eids=eids, refresh=True)
scientist4 = Scientist(["55208373700"], 2015, refresh=True)
scientist5 = Scientist(["55208373700"], 2018, period=2, refresh=True)


def test_affiliation():
    org = 'University of Essex; Universität München'
    assert_equal(scientist1.organization, org)
    org = 'École Polytechnique Fédérale de Lausanne (EPFL)'
    assert_equal(scientist2.organization, org)
    org = 'Department of Economics and Management of Innovation, '\
          'École Polytechnique Fédérale de Lausanne'
    assert_equal(scientist3.organization, org)
    org = 'École Polytechnique Fédérale de Lausanne, CEMI, CDM MTEI-GE'
    assert_equal(scientist4.organization, org)
    org = 'Max Planck Institute for Innovation and Competition'
    assert_equal(scientist5.organization, org)


def test_affiliation_id():
    assert_equal(scientist1.affiliation_id, '60001359; 60028717')
    assert_equal(scientist2.affiliation_id, '60028186')
    assert_equal(scientist3.affiliation_id, '60028186')
    assert_equal(scientist4.affiliation_id, '60028186')
    assert_equal(scientist5.affiliation_id, '60105007')


def test_active_year():
    assert_equal(scientist1.active_year, 2001)
    assert_equal(scientist2.active_year, 2017)
    assert_equal(scientist3.active_year, 2016)
    assert_equal(scientist4.active_year, 2012)
    assert_equal(scientist5.active_year, 2018)


def test_country():
    assert_equal(scientist1.country, "United Kingdom")
    assert_equal(scientist2.country, "Switzerland")
    assert_equal(scientist3.country, "Switzerland")
    assert_equal(scientist4.country, "Switzerland")
    assert_equal(scientist5.country, "Germany")


def test_citations():
    assert_true(scientist1.citations >= 47)
    assert_true(scientist2.citations >= 28)
    assert_true(scientist3.citations >= 2)
    assert_true(scientist4.citations >= 19)
    assert_true(scientist5.citations >= 44)


def test_citations_period():
    assert_equal(scientist1.citations_period, scientist1.citations)
    assert_equal(scientist2.citations_period, scientist2.citations)
    assert_equal(scientist3.citations_period, scientist3.citations)
    assert_equal(scientist4.citations_period, scientist4.citations)
    assert_equal(scientist5.citations_period, 2)


def test_coauthors():
    assert_true(isinstance(scientist1.coauthors, set))
    expected = {'7005044638', '6602701792', '35838036900', '6506756510',
                '24364642400', '6506571902', '6506426539', '6701494844',
                '11042582400', '7004064836', '7101829476'}
    assert_equal(len(scientist1.coauthors), len(expected))
    for coauth in expected:
        assert_true(coauth in scientist1.coauthors)
    assert_true(isinstance(scientist2.coauthors, set))
    expected = {'55875219200', '54929867200', '57191249971', '36617057700',
                '24464562500', '54930777900', '24781156100'}
    assert_equal(len(scientist2.coauthors), len(expected))
    for coauth in expected:
        assert_true(coauth in scientist2.coauthors)
    assert_true(isinstance(scientist3.coauthors, set))
    expected = {'36617057700', '55875219200', '54930777900', '54929867200'}
    assert_equal(len(scientist3.coauthors), len(expected))
    for coauth in expected:
        assert_true(coauth in scientist3.coauthors)
    expected = {'36617057700', '54929867200', '54930777900', '55875219200'}
    assert_equal(len(scientist3.coauthors), len(expected))
    for coauth in expected:
        assert_true(coauth in scientist5.coauthors)
    expected = {'24464562500', '24781156100', '36617057700', '54929867200',
                '54930777900', '55875219200', '57131011400', '57191249971'}
    assert_equal(len(scientist5.coauthors), len(expected))
    for coauth in expected:
        assert_true(coauth in scientist5.coauthors)


def test_coauthors_period():
    assert_equal(scientist1.coauthors_period, scientist1.coauthors)
    assert_equal(scientist2.coauthors_period, scientist2.coauthors)
    assert_equal(scientist3.coauthors_period, scientist3.coauthors)
    assert_equal(scientist4.coauthors_period, scientist4.coauthors)
    expected = {"24464562500", "24781156100", "54930777900", "55875219200",
                "57131011400", "57191249971"}
    assert_equal(len(scientist5.coauthors_period), len(expected))
    for coauth in expected:
        assert_true(coauth in scientist5.coauthors_period)


def test_fields():
    expected = [1803, 1405, 1408, 1803, 1408, 3301, 2002, 2002, 2308, 2003,
                2002, 2000, 1405, 3317, 2002, 1400, 2002, 1402, 1400, 2002,
                2200, 2308, 2002, 1405, 2000, 2002, 2003, 2002, 1400, 2002,
                1400, 1402, 2002, 3317, 1803, 1408, 1405, 1803, 1408, 2002,
                3301]
    assert_equal(scientist1.fields, expected)
    expected = [1803, 1405, 1408, 3300, 2300, 1400, 1405, 2002, 2200, 2002,
                1405, 1400, 3300, 2300, 1405, 1803, 1408]
    assert_equal(scientist2.fields, expected)
    expected = [1803, 1405, 1408, 2002, 2200, 2002, 1405, 1803, 1408]
    assert_equal(scientist3.fields, expected)


def test_first_year():
    assert_equal(scientist1.first_year, 1996)
    assert_equal(scientist2.first_year, 2012)
    assert_equal(scientist3.first_year, 2016)


def test_sources():
    received = scientist1.sources
    expected = [(17472, 'Journal of Banking and Finance'),
                (20022, 'Economic Policy'),
                (21307, 'Management Science'),
                (22900, 'Research Policy'),
                (24204, 'Review of Economics and Statistics'),
                (24389, 'Journal of Industrial Economics'),
                (26878, 'Journal of Population Economics'),
                (28994, 'Journal of Evolutionary Economics'),
                (10600153365, 'Economics of Innovation and New Technology')]
    assert_equal(len(received), len(expected))
    for e in expected:
        assert_true(e in received)
    received = scientist2.sources
    expected = [(23013, 'Industry and Innovation'),
                (21100858668, None),
                (15143, 'Regional Studies'),
                (18769, 'Applied Economics Letters'),
                (22900, 'Research Policy')]
    assert_equal(len(received), len(expected))
    for e in expected:
        assert_true(e in received)
    received = scientist3.sources
    expected = [(18769, "Applied Economics Letters"),
                (22900, "Research Policy")]
    assert_equal(len(received), len(expected))
    for e in expected:
        assert_true(e in received)


def test_sources_change():
    backup = scientist1.sources
    expected = [(14351, "Brain Research Reviews"),
                (18632, "Progress in Brain Research")]
    scientist1.sources, _ = zip(*expected)
    assert_equal(scientist1.sources, expected)
    scientist1.sources = backup


def test_main_field():
    assert_equal(scientist1.main_field, (2002, "ECON"))
    assert_equal(scientist2.main_field, (1405, "BUSI"))
    assert_equal(scientist3.main_field, (1803, "BUSI"))


def test_name():
    assert_equal(scientist1.name, "Harhoff, Dietmar")
    assert_equal(scientist2.name, "Baruffaldi, Stefano Horst")
    assert_equal(scientist3.name, "Baruffaldi, Stefano Horst")


def test_publications():
    cols = ["eid", "affiliation_country", "author_names", "author_ids",
            "author_afids", "coverDate", "authkeywords",
            "citedby_count", "description"]
    # scientist1
    received = scientist1.publications
    assert_true(len(received) >= 11)
    assert_equal(received[2].eid, '2-s2.0-0001093103')
    for p in received:
        assert_true(all(c in p._fields for c in cols))
    # scientist2
    received = scientist2.publications
    assert_true(len(received) >= 7)
    assert_equal(received[-1].eid, "2-s2.0-84866317084")
    for p in received:
        assert_true(all(c in p._fields for c in cols))
    # scientist3
    received = scientist3.publications
    assert_true(len(received) >= 2)
    received = scientist2.publications
    assert_true(len(received) >= 7)
    assert_equal(received[0].eid, "2-s2.0-85015636484")
    for p in received:
        assert_true(all(c in p._fields for c in cols))


def test_publications_period():
    assert_equal(scientist1.publications_period, scientist1.publications)
    assert_equal(scientist2.publications_period, scientist2.publications)
    assert_equal(scientist3.publications_period, scientist3.publications)
    assert_equal(scientist4.publications_period, scientist4.publications)
    assert_equal(len(scientist5.publications_period), 4)


def test_language():
    assert_equal(scientist1.language, None)
    scientist1.get_publication_languages()
    assert_equal(scientist1.language, "eng")
    scientist3.get_publication_languages()
    assert_equal(scientist3.language, "eng")



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Tests for class `Original`."""

from collections import namedtuple
from nose.tools import assert_equal, assert_true
import warnings
import pandas as pd

import sosia

refresh = False
warnings.filterwarnings("ignore")
scientist1 = sosia.Original(55208373700, 2017, cits_margin=200, refresh=refresh)
scientist2 = sosia.Original(55208373700, 2017, cits_margin=1, pub_margin=1,
                            coauth_margin=1, period=3, refresh=refresh)
affs = [60002612, 60032111, 60000765]
scientist3 = sosia.Original(55208373700, 2017, cits_margin=1, pub_margin=1,
                            coauth_margin=1, period=3, refresh=refresh,
                            search_affiliations=affs)
affs = [60010348, 60022109, 60017317]
scientist4 = sosia.Original(55208373700, 2017, cits_margin=200,
                            refresh=refresh, search_affiliations=affs)
fields = "ID name first_year num_coauthors num_publications num_citations "\
         "country language reference_sim abstract_sim"
Match = namedtuple("Match", fields)
MATCHES = [
    Match(
        ID='53164702100',
        name='Sapprasert, Koson',
        first_year=2011,
        num_coauthors=7,
        num_publications=6,
        num_citations=190,
        country='Norway',
        language='eng',
        reference_sim=0.0212,
        abstract_sim=0.1702),
    Match(
        ID='55071051800',
        name='Doldor, Elena',
        first_year=2013,
        num_coauthors=6,
        num_publications=8,
        num_citations=19,
        country='United Kingdom',
        language='eng',
        reference_sim=0.0,
        abstract_sim=0.1021),
    Match(
        ID='55317901900',
        name='Siepel, Josh',
        first_year=2013,
        num_coauthors=8,
        num_publications=7,
        num_citations=52,
        country='United Kingdom',
        language='eng',
        reference_sim=0.0079,
        abstract_sim=0.1274),
    Match(
        ID='55804519400',
        name='González, Domingo',
        first_year=2013,
        num_coauthors=6,
        num_publications=7,
        num_citations=1,
        country='Peru',
        language='eng; spa',
        reference_sim=0.0,
        abstract_sim=0.1183)]


def test_search_sources():
    scientists_list = [scientist1, scientist2, scientist3, scientist4]
    for s in scientists_list:
        s.define_search_sources()
        search_sources = s.search_sources
        assert_equal(len(search_sources), 65)
        assert_true((14726, "Technovation") in search_sources)
        assert_true((15143, "Regional Studies") in search_sources)


def test_search_sources_change():
    backup = scientist1.search_sources
    expected = [(14351, "Brain Research Reviews"),
                (18632, "Progress in Brain Research")]
    scientist1.search_sources, _ = zip(*expected)
    assert_equal(scientist1.search_sources, expected)
    scientist1.search_sources = backup


def test_search_group():
    scientist1.define_search_group(refresh=refresh)
    group = scientist1.search_group
    assert_true(isinstance(group, list))
    assert_true(365 <= len(group) <= 375)


def test_search_group_period():
    scientist2.define_search_group(ignore_first_id=True, refresh=refresh)
    group = scientist2.search_group
    assert_true(isinstance(group, list))
    assert_true(4355 <= len(group) <= 4365)


def test_search_group_stacked():
    scientist1.define_search_group(stacked=True, refresh=refresh)
    group = scientist1.search_group
    assert_true(isinstance(group, list))
    assert_true(600 <= len(group) <= 610)


def test_search_group_stacked_period():
    scientist2.define_search_group(stacked=True, ignore_first_id=True,
                                   refresh=refresh)
    group = scientist2.search_group
    assert_true(isinstance(group, list))
    assert_true(4850 <= len(group) <= 4950)


def test_search_group_stacked_period_affiliations():
    scientist3.define_search_group(stacked=True, ignore_first_id=True,
                                   refresh=refresh)
    group = scientist3.search_group
    assert_true(isinstance(group, list))
    assert_true(50 <= len(group) <= 60)


def test_search_group_stacked_affiliations():
    scientist4.define_search_group(stacked=True, ignore_first_id=True,
                                   refresh=refresh)
    group = scientist4.search_group
    assert_true(isinstance(group, list))
    assert_true(15 <= len(group) <= 22)


def test_find_matches():
    recieved = sorted(scientist1.find_matches(refresh=refresh))
    assert_equal(len(recieved), len(MATCHES))
    assert_true(isinstance(recieved, list))
    cols = ["ID", "name", "first_year", "num_coauthors", "num_publications",
            "country", "reference_sim"]
    df_r = pd.DataFrame(recieved)
    df_r["reference_sim"] = df_r["reference_sim"].round(3)
    df_m = pd.DataFrame(MATCHES)
    df_m["reference_sim"] = df_m["reference_sim"].round(3)
    pd.testing.assert_frame_equal(df_r[cols], df_m[cols])
    for e in recieved:
        assert_true(isinstance(e.abstract_sim, float))
        assert_true(0 <= e.abstract_sim <= 1)


def test_find_matches_stacked():
    recieved = scientist1.find_matches(stacked=True, refresh=refresh)
    assert_equal(len(recieved), len(MATCHES))
    assert_true(isinstance(recieved, list))
    cols = ["ID", "name", "first_year", "num_coauthors", "num_publications",
            "country", "reference_sim"]
    df_r = pd.DataFrame(sorted(recieved))
    df_m = pd.DataFrame(MATCHES)
    pd.testing.assert_frame_equal(df_r[cols], df_m[cols])
    for e in recieved:
        assert_true(isinstance(e.abstract_sim, float))
        assert_true(0 <= e.abstract_sim <= 1)


def test_find_matches_stacked():
    recieved = scientist2.find_matches(stacked=True, refresh=refresh,
                                       information=False)
    expected = [36998825200, 56049973600, 56896085200, 57188695848, 57188709931]
    assert_equal(sorted(recieved), expected)


def test_find_matches_stacked_period_affiliations():
    info = ["first_name", "surname", "first_year", "num_coauthors",
            "num_publications", "num_citations", "num_coauthors_period",
            "num_publications_period", "num_citations_period", "subjects",
            "country", "affiliation_id", "affiliation"]
    recieved = scientist3.find_matches(stacked=True, refresh=refresh,
                                       information=info)
    recieved = pd.DataFrame(recieved)
    expect_ids = ['56049973600', '56896085200', '57188695848', '57188709931']
    expect_afids = ['60000765', '60000765', '60002612', '60032111']
    assert_equal(sorted(recieved.ID.tolist()), expect_ids)
    assert_equal(sorted(recieved.affiliation_id.tolist()), expect_afids)
    

def test_find_matches_stacked_affiliations():
    recieved = scientist4.find_matches(stacked=True, refresh=refresh)
    recieved = pd.DataFrame(recieved)
    expect_m = [m for m in MATCHES if m.ID != '55804519400']
    expect_ids = [m.ID for m in expect_m]
    expect_afids = ['60010348', '60017317', '60022109']
    assert_equal(sorted(recieved.ID.tolist()), expect_ids)
    assert_equal(sorted(recieved.affiliation_id.tolist()), expect_afids)


def test_find_matches_noinfo():
    recieved = scientist1.find_matches(information=False, refresh=refresh)
    assert_equal(sorted(recieved), [int(m.ID) for m in MATCHES])
